# frozen_string_literal: true
module Hermes
  class HousingRequest
    require 'securerandom'
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::AuditableConcern
    include Hermes::Concerns::Status
    include Hermes::Concerns::ReservationHolder
    include Hermes::Concerns::SopraXmlSender
    include Hermes::Concerns::SopraXmlRequestGetter
    include Hermes::Concerns::FormValidatorConcern
    include Hermes::Concerns::EligibilityConcern

    include Mongoid::Locker
    index({locking_name: 1}, name: 'mongoid_locker_index', partial_filter_expression: {locking_name: {:$exists => true}}, expire_after_seconds: lock_timeout)
    
    hermes_tools 'revenues'
    hermes_errors 'housing_channel_error'
    
    FORM_VALIDATOR_CSV_PATH = [Hermes::Engine.root, %w[spec xml_bouchon], "housing_request_validation_sheet.csv"  ].flatten.join('/').to_s

    ENTITY_STATUS_MAP = {
      spent_on_completed_requests: [
        'Salarié logé',
        'SALARIE LOGE'
      ],
      spent_on_ongoing_requests: [
        'En cours de saisie',
        'EN COURS DE SAISIE',
        'A valider',
        'A VALIDER',
        'En cours',
        'EN COURS',
        'Dossier présenté',
        'DOSSIER PRESENTE',
        'Dossier accepté',
        'DOSSIER ACCEPTE'
      ]
    }.freeze

    REMOVE_ENTITY_STOCK_STATUS = %w[
      ANNULE
      REFUSE
    ].freeze
    
    CATEGORY_COUNTED_CHILDREN = [
      'enfant à charge',
      'garde alternée',
      'G',
      'E',
    ].freeze
    
    CHILDREN_TYPES = CATEGORY_COUNTED_CHILDREN + [
      'droit de visite',
      'D',
    ].freeze

    CATEGORY_COUNTED_DEPENDENT = CATEGORY_COUNTED_CHILDREN + [
      'P',
      'Parent',
      'A',
      'Autre'
    ].freeze
    
    CONJOINT_TYPES = [
      'Conjoint',
      'M',
      'Pacsée',
      'P',
      'Concubin (e)',
      'C',
    ].freeze

    CHANGE_SCORE_FIELDS = %w[
      entity_id
      is_active
      request_status
      created_at
      sne_created_at
      entity_validated
      bonus_points
      requester_birth_date
      priority_points_indexes
      motivations
      disability_details
      application_rejected_nb
      application_not_selected_nb
      application_taken_back_nb
      active_application_id
      consent_level
      applicant_used_for_siret
    ].freeze
    
    TYPOLOGY_FIELDS = %w[
      dependents
      applicant_ids
      unborn_children
    ].freeze
    
    ELIGIBILITY_FIELDS = %w[
      reference_tax_revenues
      monthly_revenues
      is_active
      seeked_districts
      seeked_types
      priority_points_indexes
      mutation
      mobility
      priority_public
      single_parent
      seeked_departments
      entity_validated
      disability_details
      housing_offer_ids
      main_applicant_id
      applicant_ids
      category_type
      housing_channel_keys
      unborn_children
      typology_range
      entity_id
      bordering_insee_codes
      applicants_attached_codes
    ].freeze

    REVENUES_FIELDS = %w[
      dependents
      main_applicant_id
      applicant_ids
    ].freeze
    
    ORIENTATION_BOOLS = %w[
      single_parent
      priority_public
      mutation
    ].freeze

    BOROUGHS = {
      "75056" => [ # Paris
        "75101",
        "75102",
        "75103",
        "75104",
        "75105",
        "75106",
        "75107",
        "75108",
        "75109",
        "75110",
        "75111",
        "75112",
        "75113",
        "75114",
        "75115",
        "75116",
        "75117",
        "75118",
        "75119",
        "75120",
      ],
      "69123" =>  [ # Lyon
        "69381",
        "69382",
        "69383",
        "69384",
        "69385",
        "69386",
        "69387",
        "69388",
        "69389",
      ],
      "13055" =>  [ # Marseille
        "13201",
        "13202",
        "13203",
        "13204",
        "13205",
        "13206",
        "13207",
        "13208",
        "13209",
        "13210",
        "13211",
        "13212",
        "13213",
        "13214",
        "13215",
        "13216",
      ]
    }
    
    DAYS_BEFORE_BIRTHDATE = 196
    
    YOUNG_COUPLE_THRESHOLD = 55
    APPROXIMATE_YEAR_SECONDS = 31_557_600
    
    field :form_status, type: String
    field :is_active, type: Boolean, default: false
    field :als_id, type: String
    field :attributed_housing_district, type: String
    field :attributed_housing_district_postal_code, type: String
    field :dalo, type: Boolean, default: false
    field :asked_mobility, type: Boolean
    field :valid_mobility, type: Boolean
    field :decider, type: String
    field :application_rejected_nb, type: Integer
    field :application_not_selected_nb, type: Integer
    field :application_taken_back_nb, type: Integer
    field :mobility, type: Boolean
    field :seeked_types, type: Array
    field :entity_help_required, type: Boolean
    field :contract_type, type: String
    field :registration_number, type: String
    field :entity_validated, type: Boolean, default: false
    field :entity_refused, type: Boolean, default: false
    field :priority_points_indexes, type: Array, default: []
    field :has_priority, type: Boolean, default: false
    field :total_priority_points, type: Integer
    field :nud, type: String
    field :nb_of_requests_made, type: Integer
    field :reference_tax_revenues, type: Float
    field :dependents, type: Array
    field :category_type, type: Integer
    field :single_parent, type: Boolean
    field :id_counter, type: String
    field :id_unique, type: String
    field :id_unique_associate, type: String
    field :internal_ref, type: String
    field :forcing_duplicate, type: Boolean
    field :electronical_renewal, type: Boolean
    field :ec_law_ressources, type: String
    field :is_long_delay_anormal, type: Boolean
    field :date_calculator_indicator, type: DateTime
    field :with_car_park, type: Boolean
    field :widen_request, type: Boolean
    field :housing_type_list, type: Array # Enum
    field :expulsion_judgement_date, type: Date
    field :other_reason, type: String
    field :work_contract_type, type: Array # Enum
    field :seeked_districts, type: Array, default: []
    field :seeked_district_postal_codes, type: Array, default: []
    field :seeked_district_insee_codes, type: Array, default: []
    field :seeked_departments, type: Array, default: []
    field :truly_seeked_departments, type: Array
    field :disability_details, type: Array
    field :seeked_localisation_list, type: Array
    field :profession_details, type: Hash, default: {}
    
    field :applicant_used_for_siret_firstname, type: String
    field :applicant_used_for_siret_lastname, type: String
    field :applicant_used_for_siret_fullname, type: String
    field :applicant_used_for_siret_email, type: String
    field :applicant_used_for_siret_birth_date, type: Date

    field :requester_firstname, type: String
    field :requester_lastname, type: String
    field :requester_fullname, type: String
    field :requester_email, type: String
    field :requester_birth_date, type: Date
    
    field :last_update_user_id, type: String
    field :last_update_user_last_name, type: String
    field :last_update_user_first_name, type: String
    
    field :monthly_revenues, type: Float
    field :typology_range, type: Array # Enum
    field :min_typology, type: Integer
    field :max_typology, type: Integer
    field :unborn_children, type: Array
    field :lessor_name, type: String
    field :lessor_siren, type: String
    field :social_housing_entry_date, type: DateTime
    field :social_housing_name, type: String
    field :test_created_at, type: DateTime
    field :sne_number, type: String
    field :actual_housing_details, type: Hash, default: {}
    field :postal_details, type: Hash, default: {}
    field :owned_housing_details, type: Hash, default: {}
    field :employee_id, type: String
    field :employee_start_date, type: DateTime
    field :entity_access_level, type: String
    field :employee_unit, type: String # Enum
    field :tab_state, type: String
    field :motivations, type: Array, default: []
    field :seeking_first_floor, type: Boolean
    field :seeking_elevator, type: Boolean
    field :seeked_max_rent, type: Integer
    field :seeked_epci_first_quartils, type: Array
    field :seeked_quartil_belongings, type: Array
    field :sne_created_at, type: DateTime
    field :next_sne_renewal_date, type: Date
    field :last_sne_renewal_date, type: Date
    field :next_renewal_date_set_by_sne, type: Date
    field :soon_need_to_be_renewed, type: Boolean, default: false
    field :need_to_be_renewed, type: Boolean, default: false
    field :precisions, type: String
    field :has_multiple_attributions_under_conditions, type: Boolean
    field :seeked_housing_categories, type: Array # Enum
    field :min_surface, type: Integer, default: 0
    field :priority_public, type: Boolean
    field :mutation, type: Boolean
    field :dependents_monthly_incomes, type: Hash, default: {}
    field :dependents_monthly_charges, type: Hash, default: {}
    field :valid_seeked_types, type: Array, default: []
    field :out_of_range_types, type: Array, default: []
    field :applications_nb, type: Integer, default: 0
    field :active_applications_nb, type: Integer, default: 0
    field :last_application_date, type: DateTime
    field :housing_application_ids_was, type: Array
    field :active_application_id, type: String
    field :active_application_housing_offer_id, type: String
    field :target_offer_reserving_business, type: String

    field :is_house_owner, type: Boolean
    field :nud_ids_was, type: Array
    field :cdii, type: Boolean
    
    field :siret, type: String
    field :id_cristal, type: String
    field :siren, type: String
    field :title, type: String
    field :internal_note, type: String
    field :entity_note, type: String
    # APPLICANT MODIFIED
    # serve as triggers for before_save and after_save
    
    field :superfields, type: Hash, default: {}
    
    field :applicant_ids, type: Array, default: []
    field :housing_offer_ids, type: Array, default: []
    field :documents_required, type: Integer
    field :documents_uploaded, type: Integer
    field :workplace_insee_code, type: String
    field :nud_creations, type: Hash, default: {} # Hash of form : {department1 => oldest_creation_date, department2 => oldest_creation_date, ...}
    field :housing_channel_chosen_keys, type: Array, default: []
    field :housing_channel_pending_keys, type: Array, default: []
    field :unique_numbers, type: Array, default: []
    field :housing_channel_keys, type: Array, default: ['dc']
    field :is_sne_update_available, type: Boolean
    field :last_search_on_sne, type: DateTime
    field :first_flux_30_sended, type: Boolean
    field :is_untensed_area, type: Boolean
    field :viewed_by_colaborator_offre_logement, type: Boolean
    field :backoffice_status, type: String
    field :last_entity_code, type: String
    field :applicant_used_for_siret, type: String
    field :old_entity_id, type: BSON::ObjectId
    field :consent_level, type: Integer
    field :applicant_verified, type: Boolean, default: false
    field :should_compute_required_documents, type: Boolean, default: false
    #10095 ajout de nouveaux champs cerfa v4
    field :have_monthly_income, type: Boolean
    field :disabled_person_number, type: Integer
    field :owner_info_needed, type: Boolean
    field :cerfa_version, type: String, :default => "3"

    field :profile_score, type: Integer, default: 0
    field :bonus_points, type: Integer, default: 0
    field :score_details, type: Hash, default: {
      seniority_points: 0,
      under_30_year_points: 0,
      bonus_points: 0,
      entity_validated_points: 0,
      priority_points: 0,
      professional_mobility_points: 0,
      closer_to_workplace_points: 0,
      violence_points: 0,
      handicap_points: 0,
      workplace_insee_code_points: 0,
      application_rejected_points: 0,
      application_not_selected_points: 0,
      application_taken_back_points: 0,
    }
    field :already_had_score, type: Boolean,  default: false

    field :coapplicants, type: Array
    field :striked_off_motives, type: Hash, default: {} 
    field :is_striked_off, type: Boolean, default: false

    field :cotations, type: Hash
    field :supplement_status, type: String
    field :supplement_can_apply, type: Boolean
    field :business_cotation, type: Float

    field :bordering_insee_codes, type: Array, default: []
    field :all_eligibility_codes, type: Array, default: []
    
    field :applicants_attached_codes, type: Array, default: []

    field :updated_orientation_at, type: DateTime
    
    field :used_stock_from_entity_id, type: BSON::ObjectId

    validates :main_applicant_id, presence: true
    validate :validate_coapplicants_rel_type, if: :form_completed?
    validate :custom_unique_number_unicity
    validate :check_available_points, if: :added_priority_points? 

    attr_accessor :effort_rate, :main_user_id, :already_done
    
    belongs_to(
      :main_applicant,
      class_name: 'Hermes::Applicant',
      index: true
    )
    belongs_to(
      :entity,
      class_name: 'Hermes::Entity',
      index: true
    )
    
    belongs_to :housing_request_folder, class_name: 'Hermes::HousingRequestFolder', index: true
    
    # faire un alias en main_applicant && main_applicant_id
    # has_many :applicants, class_name: 'Hermes::Applicant' # array of coapplicants
    # pour les coapplicants => il faut un moyen de stocker leur relation avec le demandeur principale
    # de type : 'connjoint', 'paxe', 'colocataire', etc ... => trouver un moyen de gerer cette relation
    # modele intermediaire / de jointure peut-être ?
    
    has_many :housing_applications, class_name: 'Hermes::HousingApplication'
    has_many :nuds, class_name: 'Hermes::Nud'
    has_many :entity_request_account, class_name: 'Hermes::EntityRequestAccount'
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    
    writer_auth_roles 'collaborateur entreprise', "collaborateur offre logement", 'superadministrateur'

    index({ als_id: 1 }, {sparse: true, background: true, unique: true})

    BASE_ELIGIBILITY_QUERY = {
      is_active: 1,
      request_status: 1,
      monthly_revenues: 1,
      min_typology: 1,
      max_typology: 1,
    }

    index(BASE_ELIGIBILITY_QUERY.merge(all_eligibility_codes: 1), {name:"all_eligibility"})
    index(BASE_ELIGIBILITY_QUERY.merge(truly_seeked_departments: 1), {name:"department"})
    index(BASE_ELIGIBILITY_QUERY.merge(seeked_district_insee_codes: 1), {name:"seeked"})
    index(BASE_ELIGIBILITY_QUERY.merge(bordering_insee_codes: 1), {name:"bordering"})

    common_indexes *%i[
      unique_numbers
      is_untensed_area
      is_active
      dalo
      has_priority
      reference_tax_revenues
      seeked_districts
      category_type
      id_counter
      seeked_district_postal_codes
      seeked_departments
      requester_firstname
      requester_lastname
      requester_email
      applicant_used_for_siret_firstname
      applicant_used_for_siret_lastname
      monthly_revenues
      typology_range
      seeking_first_floor
      seeking_elevator
      seeked_max_rent
      seeked_housing_categories
      min_surface
      housing_application_ids
      nud_ids
      applicant_ids
      main_applicant_id
      profile_score
      soon_need_to_be_renewed
      need_to_be_renewed
      last_sne_renewal_date
      next_sne_renewal_date
      applicant_used_for_siret
      cotations
      single_parent
      priority_public
      mutation
      entity_id
      entity_validated
      applicants_attached_codes
      housing_channel_keys
      truly_seeked_departments
      seeked_district_insee_codes
      bordering_insee_codes
      min_typology
      max_typology
    ]
    
    scope :actives, -> {where(is_active: true)}
    scope :not_ended, -> {where(is_active: true, :request_status.nin => ['ANNULEE', 'SAISFAITE'])}

    before_create :before_create_actions
    before_save :before_save_actions
    after_create :after_create_actions
    before_update :before_actions # , https://dev.hermes.openstack.local/admin/families/5cffa87f4ac06e526bd93295/mapping
    after_update :after_actions
    after_save :after_save_actions
    
    HousingRequest.add_status(:single, 'entity_status')
    HousingRequest.add_status(:single, 'request_status', 'INCOMPLETE')
    
    class << self
      def search_by_consent_level(level, entity_id, base_aggregation)
        matching = {'entity_forms.level' => level.to_i}
        if entity_id.present?
          matching.merge!('entity_forms.entity_id' => BSON::ObjectId(entity_id)) if entity_id.is_a? String 
          matching.merge!('entity_forms.entity_id' => { "$in" => entity_id[:$in].map {|id| BSON::ObjectId(id)}}) if entity_id[:$in] 
        end
        aggregate = [
          {
            '$lookup' => {
              from: 'hermes_housing_request_entity_forms',
              localField: '_id',
              foreignField: 'housing_request_id',
              as: 'entity_forms'
            }
          },
          {
            '$match' => matching
          }
        ]
        selector = base_aggregation.selector.except("entity_id")
        aggregate.push({'$match' => selector }) unless selector.empty?
        collection.aggregate(aggregate)
      end
    end
    
    ###############################
    #### Notification #############
    include ::Concerns::NotificationCenter::Notifiable
    
    def self.targets_selector(housing_req, opt = {})
      if opt[:main_only]
        [housing_req.main_applicant]
      else
        main_app = housing_req.main_applicant
        (Hermes::Applicant.where(:_id.in => housing_req.get_applicant_ids || []) << main_app).compact.uniq
      end
    end
    
    def form_validator_csv_path
      FORM_VALIDATOR_CSV_PATH
    end

    def self.update_renewed_booleans
      today = Date.today
      next_month = Date.today + 1.month
      Hermes::HousingRequest.where(:next_sne_renewal_date.lt => today , need_to_be_renewed: false).update_all(need_to_be_renewed: true, updated_at: DateTime.now)
      Hermes::HousingRequest.where(:next_sne_renewal_date => today..next_month, soon_need_to_be_renewed: false).update_all(soon_need_to_be_renewed: true, updated_at: DateTime.now)

      Hermes::HousingRequest.where(:next_sne_renewal_date.gte => today, need_to_be_renewed: true).update_all(need_to_be_renewed: false, updated_at: DateTime.now)
      Hermes::HousingRequest.where(:next_sne_renewal_date.lt => today , soon_need_to_be_renewed: true).update_all(soon_need_to_be_renewed: false, updated_at: DateTime.now)
      Hermes::HousingRequest.where(:next_sne_renewal_date.gt => next_month , soon_need_to_be_renewed: true).update_all(soon_need_to_be_renewed: false, updated_at: DateTime.now)
    end 

    def self.conjoint_types
      CONJOINT_TYPES
    end
    
    def self.remove_entity_stock_status
      REMOVE_ENTITY_STOCK_STATUS
    end 
    
    def notification_mailer
      ::Hermes::NotificationCenter::HermesMailer
    end
    
    def notifiable_title(key, opt = {})
      I18n.t("#{key}.title", title: id)
    end
    
    def notifiable_body(key, opt = {})
      # datas is transfered by notification.datas
      datas = opt.fetch(:datas, {})
      I18n.t("#{key}.body", number: registration_number, **datas)
    end
    
    def notifiable_path
      "https://als-hermes-int.be-idys.com/housing_requests/#{id}"
    end
    
    ###############################
    ###############################

    def is_actually_active
      is_active && request_status != "SATISFAITE"
    end 
    
    def active_application
      Hermes::HousingApplication.find_by(id: active_application_id)
    end
    
    def housing_offers
      Hermes::HousingOffer.where(:id.in => housing_offer_ids)
    end
    
    def coapplicants_raw
      Hermes::Applicant.where(:id.in => get_applicant_ids)
    end

    def coapplicants 
      self[:coapplicants].nil? ? calculate_coapplicants : self[:coapplicants].map(&:symbolize_keys)
    end 

    def remove_coapplicant(coapp_id)
      self.coapplicants = self.coapplicants.select {|coapp| coapp[:id].to_s != coapp_id.to_s } 
    end 
    
    def all_applicants_raw
      Hermes::Applicant.where(:id.in => [*get_applicant_ids, main_applicant_id.to_s ])
    end

    def coapplicants 
      self[:coapplicants].nil? ? calculate_coapplicants : self[:coapplicants].map(&:symbolize_keys)
    end 

    def remove_coapplicant(coapp_id)
      self.coapplicants = self.coapplicants.select {|coapp| coapp[:id].to_s != coapp_id.to_s } 
    end 
    
    # old system
    def calculate_coapplicants
      return [] unless applicant_ids
      
      coapps = applicant_ids&.uniq.map do |aid|
          if applicant_link = Hermes::ApplicantLink.unscoped.find_by(applicant_ids: aid, housing_request_id: id)
            {rel_type: applicant_link.type, id: aid.to_s}
          end
        end
      coapps - ["", nil]
    end

    def update_applicant_ids
      self.applicant_ids = get_applicant_ids
    end

    def get_applicant_ids
      coapplicants ? coapplicants.map{ |c| c.symbolize_keys! && c[:id] } : []
    end 

    def update_request_from_sne(user)
      xml = get_xml_from_sne(user, unique_numbers.last)
      return xml if xml.is_a? Hash

      parsed_xml = parse_xml(xml)
      update_request_and_applicant_from_xml(parsed_xml, self, user)
    end

    def is_provider_imhoweb
      nuds&.last&.provider == 'IMHOWEB'
    end 

    def update_request_supplement(user, force_update = false, insee_code = nil)
      return unless is_provider_imhoweb || force_update
      
      xml = get_supplement_xml(user, unique_numbers.last, insee_code)
      return xml if xml.is_a? Hash

      parsed_xml = parse_xml(xml)
      update_request_supplement_values_from_xml(parsed_xml, self, user)
    end

    def update_renewal_values_from_sne(user)
      xml = get_xml_from_sne(user, unique_numbers.last)
      return xml if xml.is_a? Hash

      parsed_xml = parse_xml(xml)
      update_request_renewal_values_from_xml(parsed_xml, self, user)
    end 

    def update_valid_seeked_types
      formated_range = (typology_range || [])
      valid_types = (seeked_types || []) & formated_range
      self.valid_seeked_types = valid_types.presence
      self.out_of_range_types = valid_types - formated_range | formated_range - valid_types
    end
    
    def update_requester_data
      return unless main_applicant
      
      self.requester_firstname = main_applicant.first_name
      self.requester_lastname = main_applicant.last_name
      self.requester_fullname = main_applicant.full_name
      self.requester_email = main_applicant.email
      self.requester_birth_date = main_applicant.birth_date
    end
    


    def update_active_application_requester_data
      if active_application
        data = {} 
        [
          :requester_firstname,
          :requester_lastname,
          :requester_email,
          :requester_birth_date,
          :applicant_used_for_siret_firstname,
          :applicant_used_for_siret_lastname,
          :applicant_used_for_siret_email,
          :applicant_used_for_siret_birth_date,
        ].each do |att|
          data[att] = self[att]
        end
        active_application.set(data.merge(updated_at: DateTime.now))
      end 
    end 
    
    def handle_active_request
      return unless main_applicant
      
      if is_active
        other_active_requests = main_applicant.housing_requests.actives
                                  .where(id: {:$ne => id})
        other_active_requests.update(is_active: false, request_status: 'ANNULEE', updated_at: DateTime.current)
      end
    end
    
    def main_applicant_revenues
      main_applicant&.monthly_revenues || 0
    end
    
    def coapplicants_monthly_revenues
      (coapplicants_raw || []).inject(0) do |sum, coapplicant|
        sum + coapplicant.monthly_revenues
      end
    end
    
    def dependents_monthly_revenues
      (dependents || []).inject(0) do |sum, dependent|
        dependent.to_h.symbolize_keys!
        dependent_incomes = (dependent[:monthly_incomes]&.values_fsum || 0)
        dependent_charges = (dependent[:monthly_charges]&.values_fsum || 0)
        sum + dependent_incomes - dependent_charges
      end
    end
    
    def update_total_priority_points
      self.total_priority_points = (priority_points_indexes || []).sum
    end
    
    def update_monthly_revenues
      self.monthly_revenues = main_applicant_revenues +
        coapplicants_monthly_revenues +
        dependents_monthly_revenues
    end
    
    def couple_type
      if single_parent
        'single_parent'
      elsif coapplicants.any? do |coapplicant|
        next unless coapplicant

        coapplicant.to_h.symbolize_keys!
        CONJOINT_TYPES.include?(coapplicant[:rel_type])
      end
        'couple'
      else
        'single'
      end
    end
    
    def conjoint
      conjoint_id = coapplicants.find do |coapplicant|
        next unless coapplicant

        coapplicant.to_h.symbolize_keys!
        CONJOINT_TYPES.include?(coapplicant[:rel_type])
      end&.fetch(:id, nil)
      return unless conjoint_id
      
      Hermes::Applicant.find_by(id: conjoint_id)
    end

    
    def young_couple?
      if conjoint && coapplicant_nb == 1
        births = [main_applicant, conjoint].map(&:birth_date)
        now = Time.now.utc.to_date

        age_sum = births.inject(0) do |sum, dob|
          next unless dob.respond_to?(:strftime)

          age = now.year - dob.year - ((now.month > dob.month || (now.month == dob.month && now.day >= dob.day)) ? 0 : 1)

          (sum || 0) + age
        end

        age_sum.to_i <= YOUNG_COUPLE_THRESHOLD 
      else
        false
      end
    end
    
    def handicap?
      !disability_details.blank?
    end
    
    def coapplicant_nb
      get_applicant_ids.length
    end
    
    def children_dependent_nb
      (dependents || []).select do |dependent|
        next if dependent.blank?

        dependent.to_h.symbolize_keys!
        rel_type = dependent[:rel_type]
        CHILDREN_TYPES.include?(rel_type)
      end.length
    end

    def calculate_unborn_children_nb
      return 0 unless unborn_children
      
      total = 0
      
      unborn_children.each do |unborn_child|
        next unless unborn_child.is_a?(Hash)
        unborn_child.to_h.symbolize_keys!
        if unborn_child[:anticipated_birth] &&
          DateTime.parse(unborn_child[:anticipated_birth]) - DAYS_BEFORE_BIRTHDATE.days <= DateTime.now 
            total += unborn_child[:nb_child_birth].to_i
        end 
      end
      total
    end
    
    def less_limited_dependent_nb
      dependents_nb + calculate_unborn_children_nb
    end

    def dependents_nb
      (dependents || []).select(&:present?).length 
    end 
    
    def calculate_category
      nb_people = 1 + coapplicant_nb + dependents_nb
      additional_point = (handicap? || single_parent || (young_couple? && dependents_nb == 0)) ? 1 : 0

      return nb_people + additional_point
    end
    
    def update_category
      self.category_type = main_applicant.nil? ? nil : calculate_category
    end
    
    def calculate_min_surface
      case coapplicant_nb
      when 0
        9 +
          (
          (less_limited_dependent_nb.positive? ? 1 : 0) *
            (
            7 +
              ((less_limited_dependent_nb - 1) * 9) -
              (
              (less_limited_dependent_nb > 7 ? 1 : 0) *
                (less_limited_dependent_nb - 7)
              )
            )
          )
      else
        16 + (9 * less_limited_dependent_nb)
      end
    end

    def can_have_score
      already_had_score || ['COMPLETE', 'EN ATTENTE DE PIECES'].include?(request_status) 
    end 

    def calculate_score
      return unless can_have_score

      self.score_details['seniority_points'] = calculate_seniority_points
      self.score_details['under_30_year_points'] = calculate_under_30_year_points
      self.score_details['bonus_points'] = bonus_points || 0

      self.score_details['entity_validated_points'] = calculate_entity_validated_points
      self.score_details['priority_points'] = calculate_priority_points

      self.score_details['professional_mobility_points'] = calculate_professional_mobility_points
      self.score_details['closer_to_workplace_points'] = calculate_closer_to_workplace_points 
      self.score_details['violence_points'] = calculate_violence_points

      self.score_details['workplace_insee_code_points'] = calculate_workplace_insee_code_points 
      self.score_details['handicap_points'] = calculate_handicap_points
      
      self.score_details['application_rejected_points'] = calculate_application_rejected_points 
      self.score_details['application_not_selected_points'] = calculate_application_not_selected_points 
      self.score_details['application_taken_back_points'] = calculate_application_taken_back_points 
      
      self.profile_score = score_details.values.sum
      self.already_had_score = true
    end

    def request_created_at
      sne_created_at || created_at || DateTime.now
    end 
    
    def calculate_seniority_points
      now = DateTime.now
      months = (now.year - request_created_at.year) * 12 + now.month - request_created_at.month - (now.day >= request_created_at.day ? 0 : 1)
      [12,( months)].min * 1 + (months > 24 ? 12 : [0, months - 12].max) * 2  + [0,( months - 24)].max * 3
    end

    def calculate_under_30_year_points
      return 0 unless requester_birth_date.is_a? Date
      ( requester_birth_date + 30.years >= DateTime.current.utc ) ? 70 : 0
    end

    def applicant_work_contract_type_allow_entity_score
      ["CDIP", "CDI", "APP"].include?(get_applicant_used_for_siret&.work_contract_type)
    end 

    def calculate_entity_validated_points 
      return 0 unless applicant_work_contract_type_allow_entity_score
      return 0 if [nil, 1].include?(consent_level)
      entity_validated ? 240 : 0
    end 

    def calculate_priority_points
      return 0 unless entity && entity_validated && applicant_work_contract_type_allow_entity_score
      priority_points_indexes ? priority_points_indexes.sum : 0
    end 

    def calculate_professional_mobility_points
      (motivations || []).include?('28') ? 70 : 0
    end 

    def calculate_closer_to_workplace_points
      (motivations || []).include?('19') ? 120 : 0
    end 
    
    def calculate_violence_points
      (motivations || []).include?('06') ? 40 : 0
    end 

    def calculate_workplace_insee_code_points
      return 0 if !active_application_id || !active_application.housing_offer 
      
      active_application.housing_offer.insee_code == get_applicant_used_for_siret.workplace_insee_code ? 40 : 0
    end 
    
    def calculate_handicap_points
      return 0 unless disability_details
      has_handicap = disability_details.any? do |h|
        h = h.stringify_keys
        (h['sensory_disability_kinds'] && !(h['sensory_disability_kinds'] & ['DA', 'DV']).empty?) || 
        (h['physical_disability_kinds'] && !(h['physical_disability_kinds'] & ['MS', 'MI']).empty?) ||
        h['other_disability'] == true
      end 
      has_handicap ? 40 : 0
    end 
    
    def calculate_application_rejected_points
      (application_rejected_nb || 0) * 40
    end 
    
    def calculate_application_not_selected_points
      [20, (application_not_selected_nb || 0)].min * 5
    end 
    
    def calculate_application_taken_back_points
      [20, (application_taken_back_nb || 0)].min * 5
    end 
    
    def update_applications_nb
      self.set(applications_nb: housing_applications.count, updated_at: DateTime.current)
    end
    
    def update_housing_application_ids_was
      set(housing_application_ids_was: housing_application_ids, updated_at: DateTime.current)
    end
    
    def housing_application_ids_changed?
      formated_ids = (housing_application_ids || [])
      formated_ids_was = (housing_application_ids_was || [])
      (
        (housing_application_ids && !housing_application_ids_was) &&
        (housing_application_ids_was && !housing_application_ids)
      ) ||
        (formated_ids.length != formated_ids_was.length) ||
        !(
        formated_ids.all? {|elt| formated_ids_was.include?(elt)} &&
          formated_ids_was.all? {|elt| formated_ids.include?(elt)}
        )
    end
    
    def update_update_nud_ids_was
      set(housing_application_ids_was: housing_application_ids, updated_at: DateTime.current)
    end
    
    def nud_ids_changed?
      formated_ids = (nud_ids || [])
      formated_ids_was = (nud_ids_was || [])
      (
      (nud_ids && !nud_ids_was) &&
        (nud_ids_was && !nud_ids)
      ) ||
        (formated_ids.length != formated_ids_was.length) ||
        !(
        formated_ids.all? {|elt| formated_ids_was.include?(elt)} &&
          formated_ids_was.all? {|elt| formated_ids.include?(elt)}
        )
    end
    
    def update_min_surface
      self.min_surface = calculate_min_surface
    end

    def update_tax_revenues
      self.reference_tax_revenues = nil
      minus_twos = [main_applicant.reload.dynamic_amount_minus_two] + coapplicants_raw.map { |app| app.reload.dynamic_amount_minus_two }
      if minus_twos.any? { |v| v.is_a? Float }
        self.reference_tax_revenues = (main_applicant.dynamic_amount_minus_two || 0) + coapplicants_reference_tax_revenues
      end 
    end

    def coapplicants_reference_tax_revenues
      return 0 if get_applicant_ids.blank?

      total = 0 

      get_applicant_ids.each do |id|
        if id && app = Hermes::Applicant.find(id)
          total += (app.reload.dynamic_amount_minus_two || 0)
        end
      end

      total
    end 

    def update_typology_range
      dependent_nb = less_limited_dependent_nb
      min = 0
      max = 0
      case coapplicant_nb
        when 0
          
            if dependent_nb < 2
              min = dependent_nb + 1
            else
            min = dependent_nb < 5 ? dependent_nb : 5
            end
            max = 2 + dependent_nb
          
        else
          min =  dependent_nb < 5 ? dependent_nb + 1 : 5
          if(dependent_nb == 4)
            min = 4
          end
          max =  3 + dependent_nb
        end
      self.min_typology = min
      self.max_typology = max
      self.typology_range = (min..max).map {|i| "T#{i}"}
    end
    
    def bool_filter
      ORIENTATION_BOOLS.each_with_object({}) do |bool_name, h|
        bool = send(bool_name)
        h[bool_name] = {:$ne => true} if bool.class.name != 'TrueClass'
      end
    end
    
    def reserved_filter
      if entity && entity_validated
        {
          :$and => [
            {
              :$or => [
                {reserved: true, active_reservations: entity_id.to_s},
                {reserved: {:$ne => true}}
              ]
            }
          ]
        }
      else
        {reserved: {:$ne => true}}
      end
    end
    
    def category_filter
      { 
        "ceil_category_#{category_type}".to_sym => {:$gte => reference_tax_revenues},
        "floor_category_#{category_type}".to_sym => {:$lte => reference_tax_revenues}
      }
    end
    
    def typology_filter
      {
        typology: {
          '$in': (typology_range & ['T6', 'T7', 'T8']).empty? ? typology_range : typology_range + ['T6P']
        }
      }
    end
    
    def min_revenues_filter
      { min_revenues: {:$lte => (monthly_revenues.to_f)} }
    end
    
    
    def housing_channels_filter
      {housing_channel_keys: {'$in': housing_channel_keys}}
    end

    def only_public
      reference_tax_revenues.nil? || monthly_revenues.nil? || !is_active
    end

    def township_attachement_filter
      filter = {
        :$or => [
          {housing_channel_keys: { :$ne => 'township_attachment'}},
        ]
      }
      unless applicants_attached_codes.empty?
        filter[:$or] += [{housing_channel_keys: 'township_attachment', insee_code: {:$in => applicants_attached_codes}}]
      end 
      filter
    end 
    
    def update_bordering_insee_codes
      idf_dep = ['75', '77', '78', '91', '92', '93', '94', '95']
      is_idf = !(truly_seeked_departments & idf_dep).empty?
      insee_codes = Hermes::BorderingTown.where(insee_code: {:$in => seeked_district_insee_codes}).pluck(:bordering_towns).flatten.uniq
      
      self.bordering_insee_codes = insee_codes.filter do |code|
        if is_idf && idf_dep.include?(code[0..1])
          true
        else
          size = code[0..1].to_i > 95 ? 2 : 1
          truly_seeked_departments.any? { |dep| dep == code[0..size]}
        end 
      end

      self.bordering_insee_codes = self.bordering_insee_codes - seeked_district_insee_codes
    end
    
    def calculate_eligibility_base
      return {is_public: true} if only_public
      
      
      bool_filter.deep_merge!(reserved_filter)
                 .deep_merge!(min_revenues_filter)
                 .deep_merge!(category_filter)
                 .deep_merge!(housing_channels_filter)
                 .deep_merge!(typology_filter)
                 .deep_merge!(township_attachement_filter)
                 .deep_merge!(published: true)
    end

    def calculate_eligibility_by_type(type)
      return {} if only_public

      case type
      when 'department'
        {department: {:$in => array_blank_or_missing(truly_seeked_departments) }}
      when 'seeked'
        {insee_code: {:$in => array_blank_or_missing(seeked_district_insee_codes) }}
      when 'bordering'
        {insee_code: {:$in => array_blank_or_missing(bordering_insee_codes) }}
      when 'seeked_bordering'
        {insee_code: {:$in => array_blank_or_missing((seeked_district_insee_codes + bordering_insee_codes).uniq) }}
      when 'other_from_department'
        {
          department: {:$in => array_blank_or_missing(truly_seeked_departments) },
          insee_code: {:$nin => array_blank_or_missing((seeked_district_insee_codes + bordering_insee_codes).uniq) }
        }
      end
    end

    def all_eligibility_query
      return { is_public: true} if only_public
      query = eligibility_query_base.json_to_h
      query[:$and] ||= []
      query[:$and] += [{
        :$and =>[
          :$or =>[
            eligibility_query_department.json_to_h,
            eligibility_query_seeked.json_to_h,
            eligibility_query_bordering.json_to_h,
          ]
        ]
      }]
      query
    end 

    def eligibility_find_hint
      only_public ? "is_public_1" : "eligibility_base" 
    end 

    def find_x_eligible_offer_ids(max_offers = 6, query = {}, order_by = {date_publication_start: :desc})
      offer_ids = []

      [
        'seeked',
        'bordering',
        'other_from_department',
      ].each do |type|
        break if offer_ids == max_offers
        offer_ids += eligibles_offers(type).where(query).order_by(order_by).limit(max_offers - offer_ids.length).pluck(:id)
      end

      offer_ids
    end 
    
    def eligibles_offers(type = 'department')
      query = eligibility_query(type)
      return unless query 
      Hermes::HousingOffer.where(query).hint(eligibility_find_hint) 
    end 

    
    def is_eligible_offer(o, type = 'department')
      eligibles_offers(type).pluck(:id).include?(o.id)
    end 
    
    def priority_points?
      priority_points_indexes&.length&.positive?
    end
    
    def update_has_priority
      self.has_priority = priority_points?
    end
    
    def entity_status_to_map(status)
      ENTITY_STATUS_MAP.find {|_, arr| arr.include?(status)}&.first ||
        'spent_on_other'
    end

    def used_stock_from_entity
      used_stock_from_entity_id ? Hermes::Entity.find(used_stock_from_entity_id) : entity
    end 
    
    def update_entity_stock
      entity_w = used_stock_from_entity
      entity_w = Hermes::Entity.find(entity_id_was) if !entity_id_was.blank?
      entity_status_w = entity_status_changed? ? entity_status_was : entity_status
      priority_points_indexes_sum_w = (priority_points_indexes_was || []).sum
      priority_points_indexes_sum = (priority_points_indexes || []).sum
      
      if !entity_status_w.blank? && 
        !REMOVE_ENTITY_STOCK_STATUS.include?(entity_status_w) && ( 
        entity_id_changed? || 
        priority_points_indexes_changed? || 
        entity_status_changed?
      ) 
        update_entity_stock_increment(entity_w, entity_status_w, - priority_points_indexes_sum_w)
      end

      if !entity_id.blank? && entity_validated && !REMOVE_ENTITY_STOCK_STATUS.include?(entity_status)
        update_entity_stock_increment(used_stock_from_entity, entity_status, priority_points_indexes_sum)
      end 
    end

    def update_entity_stock_increment(e, status, value)
      return unless e
      mapped = entity_status_to_map(status)
      e.reload.update(
        "#{mapped}": (e.send(mapped) || 0) + value
      )
    end
    
    def category_any_counted_children?
      return false if dependents.blank?
      
      dependents.any? do |dependent|
        dependent.to_h.symbolize_keys!
        CATEGORY_COUNTED_CHILDREN.include?(dependent[:rel_type]) || CATEGORY_COUNTED_CHILDREN.include?(dependent[:coparentality]) 
      end
    end
    
    # handle nud_creations field - Hermes::Nud dependent
    def handle_nud_creations
      # for each unique department present for associated nuds
      # set nud_creations Hash with the oldest date for the defined department
      present_departments = nuds.pluck(:department).uniq.compact
      present_departments.each do |department|
        self.nud_creations ||= {}
        self.nud_creations[department] = nuds.where(department: department)
                                           .min(:creation_date)
      end
    end
    
    def update_single_parent
      self.single_parent = get_applicant_ids.blank? && category_any_counted_children?
    end
    
    def update_seeked_location
      self.seeked_departments = []
      self.seeked_districts = []
      self.seeked_district_postal_codes = []
      self.seeked_district_insee_codes = []
      if seeked_localisation_list
        seeked_localisation_list.each do |seeked_localisation|
          check_seeked_localisation(seeked_localisation)
          if seeked_localisation.is_a? Hash
            seeked_localisation = seeked_localisation.stringify_keys
          end
          
          seeked_localisation['codesPostaux']&.each do |postal_code|
            self.seeked_district_postal_codes.push(postal_code) if postal_code.present?
          end
          
          if seeked_localisation['codeDepartement']
            self.seeked_departments.push(seeked_localisation['codeDepartement'])
          end
          if seeked_localisation['code']
            self.seeked_district_insee_codes.push(seeked_localisation['code'])
            self.seeked_district_insee_codes += BOROUGHS[seeked_localisation['code']] if BOROUGHS[seeked_localisation['code']]
          end
          if seeked_localisation['nom']
            self.seeked_districts.push(seeked_localisation['nom'])
          end
        end
      end
      
      %i[seeked_district_postal_codes seeked_departments seeked_districts].each do |arr|
        self[arr] = self[arr].uniq.sort!
      end
    end

    def update_truly_seeked_departments
      codes = seeked_district_insee_codes.empty? ? seeked_district_postal_codes : seeked_district_insee_codes
      truly_seeked_departments = codes.map do |code|
        code[0..1].to_i > 95 ? code[0..2].to_s : code[0..1].to_s
      end.concat(seeked_departments)
      corsified_codes = []
      truly_seeked_departments.each do |code|
        if ['2B', '2A', '20'].include? code 
          corsified_codes.push(['2B', '2A', '20'])
        else
          corsified_codes.push( code)
        end
      end
      self.truly_seeked_departments = corsified_codes.flatten.uniq.sort!
    end
    
    def update_housing_offers
      inter = self.housing_offer_ids = housing_applications.map do |application|
        application.housing_offer&.id.to_s
      end
      self.housing_offer_ids = inter.uniq.compact
    end
    
    def update_pending_channels!(key)
      departments = truly_seeked_departments.join(', ')
      count = truly_seeked_departments.size
      raise Hermes::Errors::HousingChannelError, I18n.t('errors.housing_channel.unavailable_for_departments', key: key, departments: departments, count: count) if !should_add_channel_key?(key)
      
      #===== TEMPORARY BYPASS VALIDATION =======
      # to reset the bypass, remove the 3 next lines + uncomment the commented lines
      self.housing_channel_keys = [] if housing_channel_keys.nil?
      raise Hermes::Errors::HousingChannelError, I18n.t('errors.housing_channel.already_added', key: key) if housing_channel_keys.include?(key)
      housing_channel_keys.push(key) unless housing_channel_keys.include?(key)
      check_channel_key_as_field(key, true)
      
      # raise Hermes::Errors::HousingChannelError, I18n.t('errors.housing_channel.already_added', key: key) if housing_channel_pending_keys.include?(key)
      # housing_channel_pending_keys.push(key)
      # housing_channel_keys.delete(key) if housing_channel_keys.include?(key)
      # check_channel_key_as_field(key, false)
      #=========================================
      self.housing_channel_chosen_keys = [] if self.housing_channel_chosen_keys.nil?
      self.housing_channel_chosen_keys.push(key) unless self.housing_channel_chosen_keys.include?(key)
      save!
    end
    
    def update_validated_channels!(key, add = true)
      
      if add
        raise Hermes::Errors::HousingChannelError, I18n.t('errors.housing_channel.already_validated', key: key) if housing_channel_keys.include?(key)
        raise Hermes::Errors::HousingChannelError, I18n.t('errors.housing_channel.not_in_pending', key: key) if !housing_channel_pending_keys.include?(key)
        housing_channel_keys.push(key)
        housing_channel_pending_keys.delete(key)
        check_channel_key_as_field(key, true)
      else
        housing_channel_keys.delete(key)
      end
      
      save!
    end
    
    def should_add_channel_key?(key)
      settings = Hermes::HousingChannelSetting.where(:department.in => truly_seeked_departments)
      return settings.select {|setting| setting.housing_channel_keys.include? key}.any? if settings.any?
      
      true
    end
    
    def check_channel_key_as_field(key, value)
      if Hermes::HousingRequest.attribute_names.include?(key) && Hermes::HousingRequest.fields[key].type == Mongoid::Boolean
        self[key.to_sym] = value
      end
    end
    
    def any_changed_for?(constant_list)
      constant_list.any? {|field| send("#{field}_changed?")}
    end
    
    def housing_differences?
      return false if housing_offer_ids.blank? && housing_application_ids.blank?
      
      formated_offer_ids = housing_offer_ids.to_array
      formated_application_ids = housing_application_ids.to_array
      echelon = formated_offer_ids.length
      return true if echelon != formated_application_ids.length
      
      (formated_offer_ids & formated_application_ids).length != echelon
    end

    def reset_pj
      housing_request_folder.required_documents.each do |rd|
        rd.set(declared: false,status: 'Pièce manquante', updated_at: DateTime.current)
      end
    end 

    def cancel_request
      self.request_status = 'ANNULEE'
      self.is_active = false
      update_applications_nb
      reset_pj
    end

    def clear_attributes_on_other_model
      main_applicant&.set(main_employer_name: nil, updated_at: DateTime.now)
    end 
    
    def update_untensed_area
      if (seeked_district_insee_codes&.length || seeked_departments&.length)
        or_condition = []
        or_condition.push({insee_code: {"$in" => seeked_district_insee_codes}}) if seeked_district_insee_codes && seeked_district_insee_codes&.length > 0
        or_condition.push({department: {"$in" => seeked_departments}}) if seeked_departments && seeked_departments&.length > 0
        if or_condition.length > 0
          self.is_untensed_area = Hermes::UntensedArea.where({"$or" => or_condition})&.length > 0
        else
          false
        end
      
      else
        self.is_untensed_area = false
      end
    end
    
    # def orientation_changes
    #   changes = ORIENTATION_BOOLS.each_with_object({}) do |orientation_name, h|
    #     if send("#{orientation_name}_changed?")
    #       h[orientation_name.to_sym] = if send(orientation_name)
    #                                      -1
    #                                    else
    #                                      1
    #                                    end
    #     end
    #   end
    #   if entity_id_changed?
    #     changes[:entities] = {
    #       entity_id_was.to_sym => -1,
    #       entity_id.to_sym => 1
    #     }
    #   end
    #   changes
    # end
    
    # def orientation_changed?
    #   orientation_changes.present?
    # end
    
    # def handle_orientation_movements
    #   return unless eligibility_query && eligibility_query != "{is_public:true}"
    
    #   HandleRequestOrientationMovementJob.perform_later(id.to_s, orientation_changes)
    # end
    def check_seeked_localisation(seeked_localisation)
      if !(seeked_localisation['nom'].is_a? String) || 
        !(seeked_localisation['code'].is_a? String) || 
        !(seeked_localisation['codesPostaux'].is_a? Array) || 
        !(seeked_localisation['codeDepartement'].is_a? String)
        errors.add(:base, "non-conform seeked_localisation_list, it should looks like [{'nom':'paris', 'code':'75113', 'codesPostaux':['75013',... },{'codeDepartement':'75'},{...},...]")
      end
    end
    
    def set_log_action_history(action)
      Hermes::LogJob.add_to_queue(
        code: 'HERMES_HISTORY_LOG',
        action: "HousingRequest #{action}",
        housing_request_id: self.id.to_s,
        current_user_id: Hermes::HousingRequest.current_user_id&.to_s
      )
      Hermes::HousingRequest.current_user_id = nil
    end
    
    #methode permettant de recuperer tout les logs d'une housing request et de les donner a voir sous une forme lisible facilement
    def get_logs(hr_id = nil)
      log_array = []
      logs = Log.where(housing_request_id: hr_id || self.id).order_by(created_at: :desc).only(:created_at, :action, :current_user_id)
      logs.each do |log|
        log_array.push({
                         date: log.created_at,
                         action: log.action,
                         user_id: log.current_user_id.to_s
                       })
        return log_array
      end
    end
    
    # cette methode permet de checker quelle est la demande qui a été vue suivant l'heure de connexion
    def security_check_request_viewed_by_other_script(created_at_value = nil, mail)
      applicant = Hermes::Applicant.find_by(email: mail)
      hr = Hermes::HousingRequest.where(:created_at.lt => created_at_value || self.created_at, entity_id: applicant&.entity_id).order_by(created_at: :desc).first
      res = {
        viewed_housing_request_id: hr._id.to_s,
        viewed_created_at: hr.created_at,
        viewed_updated_at: hr.updated_at,
        viewed_entity_id: hr&.entity&.id&.to_s,
        viewed_mail: hr.main_applicant&.email,
        viewed_first_name: hr.main_applicant&.first_name,
        viewed_last_name: hr.main_applicant&.last_name,
        requester_applicant_id: applicant&.id&.to_s,
        requester_applicant_user_id: applicant&.user_id&.to_s,
        requester_applicant_entity_id: applicant&.entity_id&.to_s,
        
      }
      return res
    end
    
    #methode permettant de verifier si la demande a été éditée par quelqu'un qui n'a pas (ou plus) les droits sur l'objet
    def check_security_anomaly(hr_id = nil)
      hr = hr_id ? Hermes::HousingRequest.find(hr_id) : self
      log_array = hr.get_logs(hr_id || self.id)
      security_violations = []
      log_array.each do |log|
        if !hr.writable_by?(User.find(log.current_user_id))
          security_violation.push(log)
        end
      end
      return security_violations
    end
    
    def security_script_from_csv_to_csv(csv_input_path, csv_output_path, with_seecurity_anomaly_check = true)
      require 'csv'
      input_data = CSV.read(csv_input_path)
      output_data = []
      input_data.each_with_index do |input_line, i|
        if i > 0
          requester_email = input_line[0].split(';')[0]
          requester_at = input_line[0].split(';')[1]
          res = security_check_request_viewed_by_other_script(requester_at, requester_email)
          res[:requester_email] = requester_email
          res[:requester_at] = requester_at
          if with_seecurity_anomaly_check
            secu_violations = check_security_anomaly(res[:viewed_housing_request_id])
            res[:security_potential_violation_nb] = secu_violations.length
            res[:security_potential_violation_detail] = secu_violations.to_json
          end
          output_data.push(res)
        end
      end
      CSV.open(csv_output_path, "wb") do |csv|
        csv << output_data.first.keys # adds the attributes name on the first line
        output_data.each do |hash|
          csv << hash.values
        end
      end
    end
    
    def validate_all_required_document
      update(form_status:"complete")
      housing_request_folder.required_documents.each do |rd|
        rd.update(declared: true)
      end
    end

    def force_validate_all_required_document
      housing_request_folder.required_documents.each do |rd|
        rd.update(status: "Pièce validée", declared: true)
      end
      housing_request_folder.reload
      self.form_status = 'complete'
    end 
    
    def before_create_actions
      handle_nud_creations if nud_ids_changed?
      update_applications_nb
      update_tax_revenues
      update_min_surface
      update_typology_range
      update_valid_seeked_types
      if entity_validated && priority_points_indexes_changed?
        update_has_priority
        update_total_priority_points
      end
      update_single_parent
      update_monthly_revenues
      update_category
      update_housing_offers
      update_eligibility_query
      self.applicant_verified = true
    end

    def manage_entity_status
      return if request_status == 'SATISFAITE' && entity_status == 'SALARIE LOGE'

      self.entity_status = 
        if entity_id.nil? 
          nil
        elsif request_status == "ANNULEE"
          "ANNULE"
        elsif entity_refused
          "REFUSE"
        elsif entity_validated && request_status == 'SATISFAITE'
          'SALARIE LOGE'
        elsif entity_validated && ['RETENUE EN CAL'].include?(active_application&.status)
          'DOSSIER ACCEPTE'
        elsif entity_validated && (
            ['TRANSMIS BAILLEUR', 'RETENUE EN CAL SOUS CONDITIONS', 'RETENUE EN CAL SOUS RESERVE', 'INSCRIPTION CAL'].include?(active_application&.status) || 
            ['VISITE PROPOSEE', 'VISITE EFFECTUEE'].include?(active_application&.visit_status) 
          )
          'DOSSIER PRESENTE'
        elsif entity_validated
          "EN COURS"
        elsif ["COMPLETE", "SATISFAITE"].include?(request_status)
          "A VALIDER"
        elsif entity_status.nil? || ["INCOMPLETE", "EN ATTENTE DE NUD", "EN ATTENTE DE PIECES"].include?(request_status)
          "EN COURS DE SAISIE"
        else
          self.entity_status
        end
    end

    def manage_housing_channel_keys
      self.housing_channel_keys += ['dc'] unless housing_channel_keys.include?('dc')
    end 

    def find_entity_id
      last_entity_code ? find_entity_id_with_entity_code : find_entity_id_with_siret
    end 

    def find_entity_id_with_entity_code
      entity_code = Hermes::EntityCode.find_by(code: last_entity_code, applicant_id: main_applicant_id)
      entity_code = Hermes::EntityCode.find_by(code: last_entity_code) unless entity_code
      entity_code.entity_id if entity_code && !entity_code.expired?
    end
    
    def find_entity_id_with_siret
      app = get_applicant_used_for_siret
      return unless app.main_employer_siret&.length == 14

      app_entity = Hermes::Entity.find_by(siret: app.main_employer_siret)
      return app_entity.id unless  app_entity.nil?

      siren_entities = Hermes::Entity.where(siren: app.main_employer_siret[0..8]) 
      siren_count = siren_entities.count
      return nil if siren_count == 0
      return siren_entities.first.id if siren_count == 1

      root_ids = siren_entities.pluck(:root_id).uniq
      
      if root_ids.length == 1
        level_siren_entities = siren_entities.map { |e| {level: e.find_level, entity: e}}.sort { |a, b|  a[:level] <=> b[:level] }
        hightest_level_entities = level_siren_entities.select { |a| a[:level] == level_siren_entities.first[:level] }       
        return hightest_level_entities.first[:entity].id if hightest_level_entities.length == 1
      else
        found_is_master =siren_entities.where(is_master: true)
        return found_is_master.first.id if found_is_master.count == 1
      end 
    end 

    def update_entity_relationship(force_reset_consent_level = true)
      self.consent_level = nil if force_reset_consent_level
      self.old_entity_id = find_entity_id
      self.entity_id = consent_level == 1 ? nil : self.old_entity_id 
      if old_entity_id_changed?
        self.priority_points_indexes = [] 
        self.entity_validated = entity ? entity.auto_request_validation : false
        self.entity_status = nil
      end 
    end

    def reset_consent_level_score
      self.priority_points_indexes = [] 
      self.entity_validated = false
    end 
    
    def update_renewal_values
      self.next_sne_renewal_date = next_renewal_date_set_by_sne || ((last_sne_renewal_date || request_created_at.to_date ) + 1.year)
      self.need_to_be_renewed = next_sne_renewal_date < Date.today 
      self.soon_need_to_be_renewed = !need_to_be_renewed && next_sne_renewal_date - 1.month <= Date.today 
    end 

    def get_applicant_used_for_siret
      Hermes::Applicant.find(applicant_used_for_siret || main_applicant_id)
    end

    def update_applicant_used_for_siret_info
      return unless appl = get_applicant_used_for_siret
      self.applicant_used_for_siret_firstname = appl.first_name
      self.applicant_used_for_siret_lastname = appl.last_name
      self.applicant_used_for_siret_fullname = appl.full_name
      self.applicant_used_for_siret_email = appl.email.blank? ? main_applicant.email : appl.email 
      self.applicant_used_for_siret_birth_date = appl.birth_date
    end

    def before_save_actions
      self.applicant_ids ||= []
      manage_housing_channel_keys
      update_housing_channel_booleans if housing_channel_keys_changed?
      update_applicant_ids
      force_validate_all_required_document if backoffice_status_changed? && backoffice_status == "COMPLETE"
      update_requester_data if is_active_changed? || main_applicant_id_changed?

      if consent_level_changed? || last_entity_code_changed? || applicant_used_for_siret_changed? || main_applicant_id_changed?
        update_entity_relationship(!consent_level_changed?)
      end  

      reset_consent_level_score if consent_level_changed? && consent_level == 1
      check_available_points if added_priority_points? 
      
      if created_at_changed? || 
        sne_created_at_changed? || 
        last_sne_renewal_date_changed? || 
        next_sne_renewal_date.nil? ||
        next_renewal_date_set_by_sne_changed?
        update_renewal_values
      end 
      update_seeked_location if seeked_localisation_list_changed?
      if seeked_departments_changed? ||
        seeked_district_postal_codes_changed? ||
        seeked_district_insee_codes_changed?
        update_truly_seeked_departments
      end
      if seeked_district_insee_codes_changed? || seeked_departments_changed?
        update_untensed_area
      end

      if main_applicant_id_changed? || 
        coapplicants_changed? || 
        (request_status_changed? && ['ELIGIBLE', 'COMPLETE'].include?(request_status))
        update_applicants_attached_codes 
      end

      self.entity_validated = false if entity_refused
      cancel_request if form_status_changed? && form_status == "cancelled"
      
      update_form_status
      update_request_status
      manage_entity_status
      calculate_score if any_changed_for?(CHANGE_SCORE_FIELDS)
      self.viewed_by_colaborator_offre_logement = false if request_status_changed?
      update_readers_writers if user_id_changed? || entity_id_changed?
      update_applicant_used_for_siret_info if applicant_used_for_siret_changed? || main_applicant_id_changed?
      update_bordering_insee_codes if seeked_district_insee_codes_changed?
      update_all_eligibility_codes if seeked_district_insee_codes_changed? || bordering_insee_codes_changed? || truly_seeked_departments_changed?
      handle_nud_creations if nud_ids_changed?
      update_applications_nb if housing_application_ids_changed?
      update_tax_revenues if main_applicant_id_changed? || coapplicants_changed?
      if seeked_types_changed? || typology_range_changed?
        update_valid_seeked_types
      end
      if entity_validated && priority_points_indexes_changed?
        update_has_priority
        update_total_priority_points
      end
      update_single_parent if dependents_changed? || coapplicants_changed?
      update_category
      if any_changed_for?(TYPOLOGY_FIELDS)
        update_typology_range
        update_min_surface
      end 
      update_monthly_revenues if any_changed_for?(REVENUES_FIELDS)
      update_housing_offers if housing_differences?
      update_eligibility_query if any_changed_for?(ELIGIBILITY_FIELDS)
      true
    end

    def before_actions
      true
    end
    
    def after_create_actions
      AlsIdUniqJob.perform_later(id.to_s)
      
      update_main_user_main_applicant
      set_log_action_history('created') if Hermes::HousingRequest.current_user_id
      HandleRequestStateJob.set(wait: 15.days).perform_later(id.to_s, 'request_creation')
      HandleRequestStateJob.set(wait: 72.hours).perform_later(id.to_s, 'required_documents')
    end

    def clean_coapplicants
      set(updated_at: DateTime.now, coapplicants: coapplicants - ["", nil])
    end 

    def custom_unique_number_unicity
      return if unique_numbers.blank? || !is_active || (!is_active_changed? && !unique_numbers_changed?)
      
      unique_numbers.each do |uniq_n|
        if Hermes::HousingRequest.where(unique_numbers: uniq_n, is_active: true, :id.ne => id).count > 0
          errors.add :unique_number, 'Ce numéro unique est déja attribué à une housing_request avec is_active=true' 
        end
      end
    end

    def update_all_eligibility_codes
      self.all_eligibility_codes = truly_seeked_departments + seeked_district_insee_codes + bordering_insee_codes
    end 

    def update_applicants_attached_codes
      apps = [main_applicant] + coapplicants_raw
      self.applicants_attached_codes = []
      apps.each { |app| self.applicants_attached_codes += app.get_applicant_attached_codes}
      self.applicants_attached_codes = self.applicants_attached_codes.uniq
    end 
    
    def update_housing_channel_booleans
      self.dalo = housing_channel_keys.include?('dalo')
    end 
    
    def validate_coapplicants_rel_type
      invalid_coap = coapplicants.select do |coap|
        next unless coap
        
        coap.to_h.symbolize_keys!
        coap[:rel_type].blank?
      end
      
      if invalid_coap.any?
        ids = invalid_coap.map do |coap|
          next unless coap
          
          coap.to_h.symbolize_keys!
          coap[:id]
        end
        self.errors.add(:coapplicants, "rel_type can't be blank for coappllicants #{ids.join(', ')}")
        return false
      end
      
      true
    end
    
    def form_completed?
      form_status == "complete"
    end

    def added_priority_points? 
      !priority_points_indexes.blank? && priority_points_indexes_was.blank?
    end 

    def check_available_points
      total = priority_points_indexes.sum
    
      return self.errors.add(:priority_points_indexes, "Not possible to add point no entity found") unless entity
      entity.update_points
      return self.used_stock_from_entity_id = entity.id if entity.available_points >= total 
      return self.errors.add(:priority_points_indexes, "No available points on the entity") if entity.is_root
      root_entity = Hermes::Entity.find(entity.root_id)
      root_entity.update_points
      return self.errors.add(:priority_points_indexes, "No available points on the entity or root") if root_entity.available_points < total 
      return self.used_stock_from_entity_id = entity.root_id
    end 

    def update_form_status
      return if form_status == 'cancelled' 
      self.form_status = 
        if truly_seeked_departments.empty? ||
          form_front_validation_errors.length > 0 ||
          form_sne_validation_errors.length > 0 
          'invalid'
        else
          'complete'
        end
    end 

    def update_request_status
      hrf = housing_request_folder&.reload
      return if ["SATISFAITE", "LOGE", "ANNULEE"].include?(request_status) || !hrf

      self.request_status = 
      if form_status != 'complete'
        "INCOMPLETE"
      elsif unique_numbers.length == 0 
        "EN ATTENTE DE NUD"
      elsif hrf.application_status == "Dossier validé"
        "COMPLETE"
      else
        "EN ATTENTE DE PIECES"
      end
    end 

    def update_main_user_main_applicant
      if main_user_id && main_user = User.find(main_user_id)
        main_user.update(main_applicant_id: main_applicant_id)
      end
    end 
    
    def create_public_shares
      lessor = active_application&.housing_offer&.lessor
      if applicated? && lessor
        Hermes::HousingRequestShare.find_or_create_by(lessor_id: lessor.id, housing_request_id: id)
      end
    end
    
    def udapte_siret_siren
      self.siren = self.siren&.slice(0, 9)
      self.lessor_siren = self.lessor_siren&.slice(0, 9)
    end
    
    def update_main_applicant_main_user
      return unless main_applicant
      
      main_applicant.main_user_id = main_user_id
      main_applicant.save
      
      self.main_user_id = nil
    end
    
    def update_housing_request_folder
      if (should_compute_required_documents ||
        main_applicant_id_changed? ||
        coapplicants_changed? ||
        motivations_changed? ||
        unborn_children_changed? ||
        dependents_changed? ||
        housing_type_list_changed? ||
        is_untensed_area_changed? ||
        disability_details_changed?)
        self.set(updated_at: DateTime.now, should_compute_required_documents: false)
        housing_request_folder.compute_required_documents
        housing_request_folder.save
      end
    end
    
    # update only
    def after_actions
      update_entity_request_account if rented?
      set_log_action_history('updated') if Hermes::HousingRequest.current_user_id
      update_housing_application_ids_was if housing_application_ids_changed?
      update_nud_ids_was if housing_application_ids_changed?
      priority_points_indexes_changed?
      handle_active_request if is_active_changed?

      # Temporary_disabled_devops_7390
      # notify_applicant
    end
    
    def rented?
      self.entity_status == 'Salarié logé' || self.entity_status == 'SALARIE LOGE'
    end
    
    def accepted?
      self.entity_status == 'Dossier accepté' || self.entity_status == 'DOSSIER ACCEPTE'
    end
    
    def satisfied_or_cancelled?
      self.entity_status == "ANNULEE" || self.entity_status == "SATISFAITE"
    end
    
    def completed?
      self.request_status == 'COMPLETE'
    end
    
    
    def applicated?
      (
        request_status === "COMPLETE" ||
        request_status === "EN ATTENTE DE PIECES" ||
        request_status === "EN ATTENTE DE NUD"
      )
    end
    
    def update_entity_request_account
      self.entity_request_account.where(request_rented: false).set(request_rented: true)
    end
    
    # WARNING : can produce loops if miss-used
    def update_applicant_for_nuds
      nuds.each do |nud|
        nud.applicant_id = main_applicant_id
        nud.save
      end
    end
    
    def update_coapplicants_user_id
     Hermes::Applicant.where(:id.in => ((applicant_ids||[]) - (applicant_ids_was||[]) || [])).each do |coapp|
          coapp.writers.push(user&._id).flatten
          coapp.readers.push(user&._id).flatten
          coapp.save
     end
    end
    
    def update_entity_fields
      #puts "updating entity fields"
      # before save avec
      if !self.entity.nil?
        self.siret = self.entity.siret
        self.id_cristal = self.entity.id_cristal
        self.title = self.entity.title.slice(0, 25)
      else
        self.siret = nil
        self.id_cristal = nil
        self.title = nil
      end
      self.save!
    end
    
    def send_flux_35
      # disable devops 13936
      # codes = []
      # housing_channel_keys&.each do |key|
      #   channel = Hermes::HousingChannel.find_by(key: key)
      #   codes.push(channel.code_ikos) if channel&.code_ikos
      # end
      # nuds.each do |nud|

      #   if nud.unique_number.length == 18
          
      #     so = Hermes::BoClient::RequestOrientationTask.new(housing_request_id: id.to_s, codes_orientation_intiale: codes, num_unique: nud.unique_number)
      #     if Rails.env != 'test'
      #       ExecuteSopraTaskJob.set(wait: 5.seconds).perform_later(so.endpoint, so.to_xml, "post", classname: "HousingRequest", id: id.to_s)
      #     else
      #       ExecuteSopraTaskJob.set(wait: 5.seconds).perform_later(so.endpoint, so.to_xml, "post", classname = "Hermes::HousingRequest", id = id.to_s)
      #     end
      #   end
      # end
    end

    def get_local_xml
      last = last_xml
      return last if last
      
      hri = Hermes::HousingRequestInitializer.find_by(housing_request: id)
      last = hri&.last_xml
      return last if last
      nil
    end

    def send_flux_30_if_needed
      return if !als_id || first_flux_30_sended || !get_local_xml

      so = Hermes::BoClient::AddRequestTask.new(housing_request_id: id)
      send_xml(so, false, 'flux_30')
    end

    def send_flux_36_if_needed
      return unless als_id_changed? || entity_id_changed? || first_flux_30_sended_changed?
      return unless als_id && entity_id && first_flux_30_sended

      ExecuteFlux36TaskJob.perform_later(id.to_s) 
    end 

    # every save
    def after_save_actions
      AlsIdUniqJob.perform_later(id.to_s) if als_id.blank?
      check_status_changed
      create_public_shares
      handle_active_request if is_active
      update_entity_stock if priority_points_indexes_changed? || entity_id_changed? || entity_status_changed?
      update_applicant_for_nuds if main_applicant_id_changed?
      update_application_business_cotation if business_cotation_changed? || active_application_id_changed?
      update_application_score if profile_score_changed? || active_application_id_changed?
      update_application_cotations if cotations_changed?
      
      update_main_applicant_main_user if main_user_id
      send_flux_35 if housing_channel_keys_changed?
      send_flux_36_if_needed
      update_seniority_point_job if sne_created_at_changed? || created_at_changed?
      clear_attributes_on_other_model if !is_active && is_active_changed?
      delete_entity_form if consent_level_changed? && consent_level.nil?
      if any_changed_for?([:requester_firstname, :requester_lastname, :requester_email, :requester_birth_date, :applicant_used_for_siret_firstname, :applicant_used_for_siret_lastname, :applicant_used_for_siret_email, :applicant_used_for_siret_birth_date])
        update_active_application_requester_data
      end 
     # update_coapplicants_user_id if coapplicants_changed?
      
      #if entity_validated_changed?
      # Hermes::BoClient::AddRequestTask.new(housing_request_id: id).execute
      #end
      
      if entity_validated_changed? && entity_validated
        # Hermes::BoClient::AddRequestTask.new(housing_request_id: id).execute
        # notify(key: NOTIFICATION_KEYS[:housing_request][:entity_validated], channels: ['email'])
      end
      
      unless housing_request_folder_id
        set(housing_request_folder_id: Hermes::HousingRequestFolder.find_or_create_by(housing_request_id: id).id, user_id: user_id, updated_at: DateTime.current)
      end
      
      if request_status == 'COMPLETE'
        if (request_status_was != 'COMPLETE') || (updated_at.year != updated_at_was&.year)
          Hermes::Accounting.create(year: Time.new.year, target: "request", object_id: self.id)
        end
      end
      handle_nuds_changed if nud_ids_changed?
      update_applicants_readers_writers if readers_changed? || writers_changed? || coapplicants_changed? || main_applicant_id_changed?
      update_housing_request_folder
      check_if_unborn_children_job
      send_flux_30_if_needed
      true
    end

    def delete_entity_form
      Hermes::HousingRequestEntityForm.find_by(housing_request_id: id)&.delete
    end 

    def update_application_cotations
      active_application&.update(updated_at: DateTime.now, cotations: cotations.dup)
    end
    
    def update_application_business_cotation
      active_application&.set(business_cotation: self.business_cotation, updated_at: DateTime.now)
    end 
    
    def update_readers_writers
      readers = writers = []
      writers += [user_id.to_s] if user_id
      if entity_id && entity
        readers += [ "g_" + entity.group_id]
        writers += [entity.supervisor_group_id, entity.manager_group_id].map {|id| "g_" + id.to_s}
      end 
      self.readers = readers
      self.writers = writers
    end 

    def update_applicants_readers_writers
      apps = [main_applicant] + coapplicants_raw
      apps.each do |a|
        a.set(
          updated_at: DateTime.now,
          readers: self.readers,
          writers: self.writers
        )
      end
    end 

    def update_readers_writers
      readers = writers = []
      writers += [user_id.to_s] if user_id
      if entity_id && entity
        readers += [ "g_" + entity.group_id]
        writers += [entity.supervisor_group_id, entity.manager_group_id].map {|id| "g_" + id.to_s}
      end 
      self.readers = readers
      self.writers = writers
    end 

    def update_applicants_readers_writers
      apps = [main_applicant] + coapplicants_raw
      apps.each do |a|
        a.set(
          updated_at: DateTime.now,
          readers: self.readers,
          writers: self.writers
        )
      end
    end
      
    def update_application_score
      active_application&.set(profile_score: profile_score, updated_at: DateTime.now)
    end

    def calculate_next_seniority_updated_date
      request_created_at
      today = DateTime.now
      today += 1.month if request_created_at.day <= today.day

      date  = today.end_of_month.day >= request_created_at.day ? today.change(day: request_created_at.day)  : today.end_of_month
      (date + 1.days).utc.beginning_of_day
    end 
    

    def update_seniority_point_job
      date = calculate_next_seniority_updated_date
      UpdateSeniorityPointJob.set(wait_until: date + 1.hours).perform_later(id.to_s, date.to_time.to_i)
    end 

    def handle_nuds_changed
      set(unique_numbers: nuds.pluck(:unique_number).uniq, updated_at: DateTime.current)
    end

    def check_if_unborn_children_job
      return unless unborn_children
      
      unborn_children.each do |unborn_child|
        next unless unborn_child.is_a?(Hash)
        unborn_child.to_h.symbolize_keys!

        if unborn_child[:anticipated_birth] &&
          DateTime.parse(unborn_child[:anticipated_birth]) - DAYS_BEFORE_BIRTHDATE.days > DateTime.now 
          UnbornChildrenJob.set(wait_until: DateTime.parse(unborn_child[:anticipated_birth]) - (DAYS_BEFORE_BIRTHDATE - 1).days ).perform_later(self.id.to_s)
        end 
      end
    end

    
    def notify_applicant
      if entity_status_changed? && accepted?
        # notify(key: NOTIFICATION_KEYS[:housing_request][:entity_accepted], channels: ['email'], main_only: true)
      end
      
      if request_status_changed?
        # notify(key: NOTIFICATION_KEYS[:housing_request][:status_change], channels: ['email'], datas: { status: request_status })
        
        # No channels, just a front notif
        notify(key: NOTIFICATION_KEYS[:housing_request][:request_completed], channels: [], main_only: true) if completed?
      end
      
      if Hermes::HousingRequest.current_user_id && Hermes::HousingRequest.current_user_id === main_applicant.main_user_id
        notify(key: NOTIFICATION_KEYS[:housing_request][:applicant_edit], channels: [], main_only: true)
      end
    end

    def update_request_orientation(xml)
      begin 
        parsed_xml = parse_xml(xml)
        if parsed_xml['demandeOrientationsReponse']['erreurs']
          Hermes::LogJob.add_to_queue(level: 'error', housing_request_id: self.id.to_s, code: "UPDATE_ORIENTATION_ERROR", action: "error sopra", data: xml );
          return 
        end 
        self.housing_channel_keys = []
        
        if parsed_xml['demandeOrientationsReponse']['liste_orientations']
          keys = parsed_xml['demandeOrientationsReponse']['liste_orientations']['code_orientation']
          keys = keys.is_a?(Array) ? keys : [keys]

          keys.each do |key|
            update_pending_channels!(key)
          end
        end
        
        self.updated_orientation_at = DateTime.current
        save!

      rescue Exception => e
        Hermes::LogJob.add_to_queue(level: 'error', housing_request_id: self.id.to_s, code: "UPDATE_ORIENTATION_ERROR", action: "undefinned error", data: e.to_s );
      end 
    end 
  end
end
