# frozen_string_literal: true

# Legende
# Les script a plusieurs parties indique que les bash (Batch dans l'app)
# peuvent s'initier avec differents types de source
# Le besoin est donc d'isoler la logique associee au modele actuel
# ne pas oublier d'indexer les champs simples

module Hermes
  # Lessor model
  class Lessor < Hermes::Entity
    hermes_tools 'handle_dso'

    include ::Concerns::NotificationCenter::Targetable

    field :api_id, type: String
    field :rpls_id, type: String
    field :affiliate, type: String
    
    has_one :contact_group, class_name: 'Hermes::ContactGroup'
    
    #TODO: le setteur(ou le getteur)de lessor_objective ne fonctionne pas
    field :lessor_objective_ids, type: Array, default: []
    has_many :lessor_objectives, class_name: 'Hermes::LessorObjective'
    has_many :agencies, class_name: 'Hermes::Entity', foreign_key: :id, primary_key: :lessor_objective_ids
    
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    writer_auth_roles 'superadministrateur'
    owner_field :user_id
    
    after_save :after_actions
    
    Lessor.add_status(:single)
    
    def instance_values(instance)
      instance.send(case instance.class.name.demodulize
                    when 'Tagset'
                      'superfields'
                    when 'HousingOffer'
                      'attributes'
                    end)
    end
    
    def lessor_objective_ids=(arr)
      objective_ids_to_unlink = lessor_objective_ids - (arr || [])
      now = DateTime.current
      Hermes::LessorObjective.where(:id.in => objective_ids_to_unlink).set(lessor_id: nil, update_at: now)
      Hermes::LessorObjective.where(:id.in => arr).set(lessor_id: id, updated_at: now)
      super(arr)
    end

    def is_lessor
      true
    end
    
    def lessor_objective_ids
      lessor_objectives.pluck(:id)
    end
    
    def current_operation_names
      Hermes::HousingOffer.where(lessor_id: self.id).distinct(:operation_name)
    end

    def update_target_objectives_on_offers
      
      Hermes::HousingOffer.where(lessor_id: self.id.to_s).each do |housing_offer|
        if !Hermes::HousingOffer.rented_statuses.include?(housing_offer.lessor_status)
          housing_offer.target_objectives = get_target_objectives_for(housing_offer)
          housing_offer.save
        end
      end
    end
    
    def handle_objectives_for(housing_offer)
      housing_offer&.reload&.target_objectives.each do |obj|
        obj&.reload&.increment_realized_for(housing_offer.typology, housing_offer)
      end
    end
    
    def get_target_objectives_for(housing_offer)
      objectives_res = []
      offer_fields = instance_values(housing_offer)
      query = offer_fields.select do |key, val|
        filter_included = Hermes::LessorObjective.filter_keys.include?(key.to_sym)
        valid_val = val.present? || val.is_a?(FalseClass)
        filter_included && valid_val
      end
      
      unsorted_objectives = lessor_objectives.actives.where(query).where("$or" => [{present_typologies: housing_offer.typology}, {present_typologies: "all"}]).entries
      by_activation_date = []
      unsorted_objectives.each do |obj|
        found = false
        by_activation_date.each_with_index do |by_date, i|
          if by_date[:activation_date] == obj.activation_date
            by_activation_date[i][:unsorted_objectives].push(obj)
            found = true
          end
        end
        if (!found)
          by_activation_date.push({
                                    activation_date: obj.activation_date,
                                    unsorted_objectives: [obj]
                                  })
        end
      end
      
      by_activation_date.each do |by_date|
        already_incremented = false
        by_date[:unsorted_objectives].sort_by {|obj| get_score_sort(obj, housing_offer)}.each do |objective|
          unless already_incremented
            if is_to_increment(objective, housing_offer)
              objectives_res.push(objective)
              # objective.increment_realized_for(offer_fields['typology'], housing_offer)
              already_incremented = true
            end
          end
        end
      end
      return objectives_res
    end
    
    def get_score_sort(obj, offer)
      score = 100
      [:department,
       :district,
       :threshold_type,
       :priority_public,
       :new_offer,
       :operation_name].each do |key|
        if obj.send(key) == offer.send(key)
          score -= 1
        elsif !obj.send(key)
          score -= 0.01
        end
      
      end
      return score
    end
    
    def is_to_increment(objective, offer)
      # Détection du bon objectif :
      #SI (logement.operation = objectif.operation ET objectif.operation != « indéfini »)) ALORS
      if (offer.operation_name === objective.operation_name && objective.operation_name != nil)
        return priority_public_cond(objective, offer)
        #SINON SI (objectif.operation = « indéfini »)) ALORS
      elsif (objective.operation_name == nil)
        return priority_public_cond(objective, offer)
        #SINON le logement n’a pas d’objectif
      else
        return false
      end
    end
    
    def geo_cond(objective, offer)
      #SI (objectif.zoneGeo est une ville  ET logement.zoneGeo = objectif.zoneGeo) ALORS
      if (offer.district === objective.district && objective.district != nil)
        #le logement correspond à cet objectif
        return true
        #SINON SI (objectif.zoneGeo est un départ.  ET logement.zoneGeo IN objectif.zoneGeo) ALORS
      elsif (objective.district == nil && offer.department === objective.department && objective.department != nil)
        #le logement correspond à cet objectif
        return true
        #SINON le logement n’a pas d’objectif
      else
        return false
      end
    end
    
    def threshold_type_cond(objective, offer)
      #SI (logement.norme = objectif.norme ET objectif.norme != « indéfini ») ALORS
      if (offer.threshold_type === objective.threshold_type && objective.threshold_type != nil)
        return geo_cond(objective, offer)
        #SINON SI (objectif. norme = « indéfini »)) ALORS
      elsif (objective.threshold_type == nil)
        return  geo_cond(objective, offer)
        #SINON le logement n’a pas d’objectif
      else
        return false
      end
    end
    
    def new_offer_cond(objective, offer)
      #SI (logement.miseHabitation = objectif.miseHabitation ET objectif. miseHabitation != « indéfini ») ALORS
      if (offer.new_offer === objective.new_offer && !!objective.new_offer != nil)
        return threshold_type_cond(objective, offer)
        #SINON SI (objectif. miseHabitation = « indéfini »)) ALORS
      elsif (objective.new_offer == nil)
        return threshold_type_cond(objective, offer)
        #SINON le logement n’a pas d’objectif
      else
        return false
      end
    end
    
    def priority_public_cond(objective, offer)
      #SI (logement.PP = objectif.PP ET objectif.PP != « indéfini ») ALORS
      if (offer.priority_public === objective.priority_public && objective.priority_public != nil)
        return new_offer_cond(objective, offer)
        #SINON SI (objectif.PP = « indéfini »)) ALORS
      elsif (objective.priority_public == nil)
        return new_offer_cond(objective, offer)
        #SINON le logement n’a pas d’objectif
      else
        return false
      end
    end
    
    def sum_objectives_for(objectives, realized_or_target)
      return unless objectives.present?
      
      objectives.inject(0) do |sum, objective|
        sum + objective.sum_by_typologies_for(realized_or_target)
      end
    end
    
    def housing_actual_flow_number
      sum_objectives_for(lessor_objectives.actives, 'realized')
    end
    
    def housing_targeted_flow_number
      sum_objectives_for(lessor_objectives.actives, 'target')
    end
    
    def new_housing_actual_flow_number
      sum_objectives_for(lessor_objectives.actives.with_new, 'realized')
    end
    
    def new_housing_targeted_flow_number
      sum_objectives_for(lessor_objectives.actives.with_new, 'target')
    end
    
    def after_actions
      check_status_changed
      if user_id.present? && user_id_changed?
        user = ::User.find(user_id)
        user.update!(lessor_id: self.id)
      end
    end
  end
end
