# frozen_string_literal: true
require 'rails_helper'
require 'vcr_config'


module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

  describe Hermes::Api::Dmo::EligibleCountOffersController, type: :api do
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper

    let!(:superadministrateur) {create(:user, auth_roles: ['superadministrateur'])}
    let!(:demandeur) {create(:user, auth_roles: ['demandeur'])}

    let!(:offer) {create(:housing_offer, offer_status: 'PUBLIE')}
    let!(:eligible_count_offer) {create(:eligible_count_offer, housing_offer: offer)}
    let!(:eligibility_count_offer2) {create(:eligible_count_offer, housing_offer: offer)}


    describe 'GET #index' do
      it 'return 401 if no access' do
        authenticated_header(request, demandeur)
        get 'index'
        expect(response).to have_http_status(401)
      end 

      it 'return 200 if access' do
        authenticated_header(request, superadministrateur)
        get 'index'
        expect(response).to have_http_status(200)
        expect(json).not_to be_empty
        expect(json['data'].length).to eq(2)
      end 
    end 

    describe 'GET #show' do
      it 'return 401 if no access' do
        authenticated_header(request, demandeur)
        get 'show', id: eligible_count_offer.id
        expect(response).to have_http_status(401)
      end 

      it 'return 200 if access' do
        authenticated_header(request, superadministrateur)
        get 'show', id: eligible_count_offer.id
        expect(response).to have_http_status(200)
        expect(json['data']['attributes']['housing_offer_id']).to eq(offer.id.to_s)
        expect(json['data']['attributes']['created_at']).to include(Date.today.to_s)
        expect(json['data']['attributes']['offer_status']).to eq('PUBLIE')
        expect(json['data']['attributes']['reserved_entitites_eligible_request_nb']).to eq({})
      end 
    end 
  end 
 
  require_relative '../../swagger_helper'

  describe Hermes::Api::Dmo::EligibleCountOffersController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
    TAGS = 'API_DMO_ELIGIBILITY_COUNT_OFFERS'
    include JwtAuthHelper
    extend HermesSwaggerJsonParams
    include SwaggerJsonAttributes
    let!(:superadministrateur) {create(:user, auth_roles: ['superadministrateur'])}


    path '/api/dmo/eligible_count_offers' do
      get 'Retrieves all eligible_count_offers' do
        tags TAGS
        produces 'application/json'
  
        Hermes::Api::Dmo::EligibleCountOffersController.acceptable_params.sort.each do |params_name, options|
          parameter name: params_name.to_s, in: options[:in] || :query, type: options[:type], required: options[:required] || false, description: options[:description] || ''
        end
  
        security [Bearer: {}]
        paginator
  
        response '401', 'unauthorized' do
          let(:Authorization) { "Bearer" }
  
          run_test!
        end 
  
        response '200', 'eligible_count_offers found' do
          let(:Authorization) {"Bearer #{token_for(superadministrateur)}"}
          let!(:offer) {create(:housing_offer, offer_status: 'PUBLIE')}
          let!(:eligible_count_offer) {create_list(:eligible_count_offer, 5)}

          include_context 'with integration test'
          include_context 'valid jsonapi collection' do
            let(:expected_collection) {eligible_count_offer}
          end
        end
      end
    end

    path '/api/dmo/eligible_count_offers/{id}' do
      get 'Retrieves a single eligible_count_offers by id' do
        tags tags
        produces 'application/json'
        parameter name: :id, in: :path, type: :string
        security [Bearer: {}]
        let!(:offer) {create(:housing_offer, offer_status: 'PUBLIE')}
        let!(:eligible_count_offer) {create(:eligible_count_offer, housing_offer: offer)}
        let(:id) {eligible_count_offer.id}

        response '401', 'unauthorized' do
          let(:Authorization) { "Bearer" }
          run_test!
        end 
  
        response '404', 'not found' do
          let(:Authorization) {"Bearer #{token_for(superadministrateur)}"}
          let(:id) { 1 }
          run_test!
        end 
  
        response '200', 'eligible_count_offer found' do
          let(:Authorization) {"Bearer #{token_for(superadministrateur)}"}
  
          include_context 'with integration test'
          include_context 'valid jsonapi ressource' do
            let(:expected_ressource) {eligible_count_offer}
          end
        end
      end
    end
  end
end