# frozen_string_literal: true

require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

describe Hermes::Api::Dmo::EntityReservationsController, type: :api do
  routes {Hermes::Engine.routes}
  include Devise::Test::ControllerHelpers
  include JwtAuthHelper

  let!(:responsible) {create(:user)}
  let!(:entity1) {create(:entity, siret: '00000000000001', user: responsible )}
  let!(:housing_offer1) {create(:housing_offer)}
  let!(:housing_offer2) {create(:housing_offer)}
  let!(:entity_reservation1) {create(:entity_reservation, 
    entity_id: entity1.id.to_s, 
    housing_offer_id: housing_offer1.id.to_s,
    from: Date.today - 30.days,
    to: Date.today + 30.days,
    name: 'entity_reservation1'
  )}


  describe 'GET /entity_reservations' do
    before :each do
      authenticated_header(request, responsible)
      get 'index'
    end

    it 'returns status code 200' do
      expect(response).to have_http_status(200)
      expect(json['data'].map{ |ev| ev['id'] }).to include(entity_reservation1.id.to_s)
    end
  end

  describe 'POST /entity_reservations' do
    let!(:params) { 
      {
        data: {
          attributes: {
            entity_id: entity1.id.to_s,
            housing_offer_id: housing_offer1.id.to_s,
            from: Date.today - 30.days,
            to: Date.today + 30.days,
          }
        }
      }
    }
    describe 'Create entity_reservation without minimal valid datas' do
      it 'returns status code 422' do
        authenticated_header(request, responsible)
        params[:data][:attributes][:entity_id] = nil
        post 'create',params
        expect(response).to have_http_status(422)
      end
    end
    
    describe 'Create entity_reservation with minimal valid datas' do
      it 'returns status code 201' do
        authenticated_header(request, responsible)
        params[:data][:attributes][:housing_offer_id] = housing_offer2.id.to_s
        post 'create',params
        expect(response).to have_http_status(201)
        expect(json['data']['id']).not_to eq(entity_reservation1.id.to_s)
      end
    end

    describe 'update existing reservation' do 
      it 'returns status code 201' do
        authenticated_header(request, responsible)
        post 'create', params
        expect(response).to have_http_status(201)
        expect(json['data']['id']).to eq(entity_reservation1.id.to_s)
      end
    end 
  end

  describe 'PUT /entity_reservations/:if' do
    let!(:entity_reservation2) { 
      create(:entity_reservation, 
        entity_id: entity1.id.to_s, 
        housing_offer_id: housing_offer2.id.to_s,
        from: Date.today - 30.days,
        to: Date.today + 30.days,
        name: 'reservation2'
      )
    }
    let!(:params) { 
      {
        id: entity_reservation1.id.to_s,
        data: {
          attributes: {
            entity_id: entity1.id.to_s,
            housing_offer_id: housing_offer1.id.to_s,
            from: Date.today - 28.days,
            to: Date.today + 30.days,
          }
        }
      }
    }

    it 'update from' do
      authenticated_header(request, responsible)
      post 'update', params
      expect(response).to have_http_status(200)
      expect(json['data']['id']).to eq(entity_reservation1.id.to_s)
      expect(entity_reservation1.reload.from).to eq(Date.today - 28.days)
    end 

    it 'when changing housing_offer_id' do 
      authenticated_header(request, responsible)
      params[:data][:attributes][:housing_offer_id] = housing_offer2.id.to_s

      post 'update', params
      expect(response).to have_http_status(200)
      expect(json['data']['id']).to eq(entity_reservation2.id.to_s)
      expect(entity_reservation2.reload.from).to eq(Date.today - 28.days)
    end 
  end 
end

require_relative '../../swagger_helper'

describe Hermes::Api::Dmo::EntityReservationsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
  tags = 'API_DMO_ENTITY_RESERVATIONS'
  include JwtAuthHelper
  extend HermesSwaggerJsonParams
  include SwaggerJsonAttributes
  let!(:admin) {create(:administrator)}
  let!(:offer) {create(:housing_offer)}
  let!(:entity) {create(:entity)}

  path '/api/dmo/entity_reservations' do
    get 'Retrieves all entity reservations' do
      tags tags
      produces 'application/json'

      Hermes::Api::Dmo::EntityReservationsController.acceptable_params.sort.each do |params_name, options|
        parameter name: params_name.to_s, in: options[:in] || :query, type: options[:type], required: options[:required] || false, description: options[:description] || ''
      end

      security [Bearer: {}]
      paginator

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }

        run_test!
      end 

      response '200', 'entity reservation found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let!(:entity_reservations) { create_list(:entity_reservation, 5, entity: entity, housing_offer: offer, from: DateTime.current, to: DateTime.current, user: admin) }

        include_context 'with integration test'
        include_context 'valid jsonapi collection' do
          let(:expected_collection) { entity_reservations }
        end
      end
    end
  end

  path '/api/dmo/entity_reservations' do
    post 'Create entity reservation' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::EntityReservationsController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )

      let(:params) do
        jsonapi_attributes(
          {
            entity_id: entity.id.to_s,
            housing_offer_id: offer.id.to_s,
            from: DateTime.current,
            to: DateTime.current,
            user: admin,
          }
        )
      end

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }

        run_test!
      end 

      response '201', 'entity reservation created' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }

        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'

        after do |_example|
          json = JSON.parse(response.body)

          expect(json['data']).to be_a Hash
          expect(json['data']['id']).to_not be nil
          expect(json['data']['type']).to eq 'entity_reservations'
        end
      end
    end
  end

  path '/api/dmo/entity_reservations/{id}' do
    put 'Update a single entity reservation by id' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::EntityReservationsController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )
      let!(:entity_reservation) { create(:entity_reservation, entity: entity, housing_offer: offer, from: DateTime.current, to: DateTime.current, user: admin) }
      let!(:entity2) { create(:entity) }
      let(:id) { entity_reservation.id.to_s }
      let(:params) do
        jsonapi_attributes(
          {
            entity_id: entity2.id.to_s,
            housing_offer_id: offer.id.to_s,
            from: DateTime.current,
            to: DateTime.current,
            user: admin,
          }
        )
      end
      
      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }

        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:id) { 1 }

        run_test!
      end 

      response '200', 'entity reservation updated' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }

        run_test!

        after do |_example|
          json = JSON.parse(response.body)

          data = json['data']
          expect(data).to be_a Hash
          expect(data['id']).to_not be nil
          attributes = data['attributes']
          expect(attributes['entity_id']).to eq entity2.id.to_s
        end
      end
    end
  end

  path '/api/dmo/entity_reservations/{id}' do
    delete 'Delete a single entity reservation by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      let!(:entity_reservation) { create(:entity_reservation, entity: entity, housing_offer: offer, from: DateTime.current, to: DateTime.current, user: admin) }
      let(:id) { entity_reservation.id }


      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }

        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:id) { 1 }

        run_test!
      end 

      response '204', 'entity reservations deleted' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        run_test!
      end
    end
  end
end
