require 'rails_helper'
require 'json'
require 'vcr_config'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
  
describe Hermes::Api::Dmo::HousingRequestsController, :type => :api do
  routes {Hermes::Engine.routes}
  include Devise::Test::ControllerHelpers
  include JwtAuthHelper
  
  check_mutate_params(
    model: Hermes::HousingRequest, 
    mutate_params: Hermes::Api::Dmo::HousingRequestsController.mutate_params, 
    factory: :housing_request,
    other_factory:[
      {factory: :applicant, id: 'app1'},
      {factory: :applicant, id: 'app2'},
    ],
    min_params: {
      main_applicant_id: 'app1',
    },
    not_checked_keys: [
      :entity_validated,
      :entity_refused,
      :form_status
    ]
  )

  let!(:responsible) {create(:user, roles: ['Administrator'])}
  let!(:superadministrateur) {create(:user, auth_roles: ['superadministrateur'])}
  let!(:user) {create(:user)}
  let!(:applicant) {create(:applicant)}
  
  # common CRUD
  describe 'POST /housing_requests' do
    describe 'Create housing_request with minimal valid datas' do
      before :each do
        authenticated_header(request, responsible)
        post 'create', {
          data: {
            attributes: {
              main_applicant_id: applicant.id.to_s
            }
          }
        }
      end
      it 'returns status code 201' do
        expect(response).to have_http_status(201)
      end
    end
    
    # main_applicant_id:
    describe 'post request with main_applicant_id' do
      let!(:applicant1) {create(:applicant, user_id: user.id)}
      
      before :each do
        authenticated_header(request, responsible)
        post 'create', {
          data: {
            attributes: {
              main_applicant_id: applicant1.id.to_s,
              
            }
          }
        }
      end
      it 'returns status code 201 and the value appears in json' do
        expect(response).to have_http_status(201)
        expect(json['data']['attributes']['main_applicant_id']).to eq(applicant1.id.to_s)
      end
    end
    
    # coapplicants: 'Array'
    describe 'post request with coapplicants' do
      let!(:applicant0) {create(:applicant, user_id: user.id, email: "coucou@coucou.com")}
      let!(:applicant1) {create(:applicant, user_id: user.id)}
      let!(:applicant2) {create(:applicant, user_id: user.id)}
      
      $request_id = nil # to store global value
      
      before :each do
        applicant2.set(email: nil)
        authenticated_header(request, responsible)
        post 'create', {
          data: {
            attributes: {
              main_applicant_id: applicant0.id.to_s,
              form_status: "coco",
              coapplicants: [{'rel_type' => 'P', 'id' => applicant1.id.to_s}, {'rel_type' => 'P', 'id' => applicant2.id.to_s}],
              applicant_used_for_siret: applicant2.id.to_s
            }
          }
        }
      end
      
      it 'returns status code 201 and the value appears in json' do
        $request_id = json['data']['id']
        expect(response).to have_http_status(201)
        expect(json['data']['attributes']['applicant_ids']).to include(applicant1.id.to_s).and(include(applicant2.id.to_s))
        expect(json['data']['attributes']['coapplicants']).to eq([{'rel_type' => 'P', 'id' => applicant1.id.to_s}, {'rel_type' => 'P', 'id' => applicant2.id.to_s}])
        expect(json['data']['attributes']['applicant_used_for_siret_firstname']).to eq(applicant2.first_name)
        expect(json['data']['attributes']['applicant_used_for_siret_email']).to eq(applicant0.email)
        expect(Hermes::HousingRequest.find($request_id).applicant_used_for_siret_firstname).to eq(applicant2.first_name)
      end
      
      describe 'destroy previous applicants and set with new' do
        let!(:applicant3) {create(:applicant, user_id: user.id)}
        before :each do
          $request_id = json['data']['id']
          # (1..2).each {|i| send("applicant#{i}").destroy}
          authenticated_header(request, responsible)
          put 'update', {
            id: $request_id,
            data: {
              attributes: {
                coapplicants: [
                  {'rel_type' => 'P', 'id' => applicant1.id.to_s}
                ]
              }
            }
          }
        end
        
        it 'returns 200' do
          expect(response).to have_http_status(200)
        end
        
        it 'returns valid json' do
          expect(json['data']).to_not be_empty
          expect(json['data']['id']).to eq $request_id
          attributes = json['data']['attributes']
          expect(attributes['coapplicants'].length).to be 1
          expect(attributes['coapplicants'].first['rel_type']).to eq 'P'
          expect(attributes['coapplicants'].first['id']).to eq applicant1.id.to_s
        end
      end
      describe 'destroy previous applicants and set with new' do
        let!(:applicant3) {create(:applicant, user_id: user.id)}
        
        before :each do
          $request_id = json['data']['id']
          hr = Hermes::HousingRequest.find_by(id: json['data']['id'])
          applicant1.destroy
          applicant2.destroy
          applicant3.destroy
          
          authenticated_header(request, responsible)
          get 'show', {id: $request_id}
        end
        
        it 'returns 200' do
          expect(response).to have_http_status(200)
        end
        
        it 'returns valid json' do
          expect(json['data']).to_not be_empty
          expect(json['data']['id']).to eq $request_id
          attributes = json['data']['attributes']
          expect(attributes['coapplicants'].length).to be 0
        end
      end
    end
  end
  
  describe 'situation faille secu' do
    let!(:user1) {create(:user, auth_roles: ['demandeur'])}
    let!(:user2) {create(:user, auth_roles: ['demandeur'])}
    let!(:entity1) {create(:entity, siret: '00000000000001')}
    let!(:applicant1) {create(:applicant, user_id: user1.id, main_employer_siret: entity1.siret)}
    let!(:applicant2) {create(:applicant, user_id: user2.id, main_employer_siret: entity1.siret)}
    
    let!(:housing_request1) {create(:housing_request,
                                    documents_required: 50,
                                    main_applicant_id: applicant1.id.to_s,
                                    user_id: user1.id,
                                    main_user_id: user1.id,
                                    is_active: true)}
    let!(:housing_request2) {create(:housing_request,
                                    documents_required: 50,
                                    main_applicant_id: applicant2.id.to_s,
                                    user_id: user2.id,
                                    main_user_id: user2.id,
                                    is_active: true)}
    
    describe "as user 1" do
      before :each do
        authenticated_header(request, user1)
        get 'show', {id: "active"}
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
        expect(json['data']['id']).to eq housing_request1.id.to_s
      end
    end
    
    describe "as user 2" do
      before :each do
        authenticated_header(request, user2)
        get 'show', {id: "active"}
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
        expect(json['data']['id']).to eq housing_request2.id.to_s
      end
    end
  end
  
  describe 'POST /housing_requests/search' do
    let!(:entities) {create_list(:entity, 10)}
    let!(:entity) {create(:entity, siret: '12345678901234')}
    let!(:applicant) {create(:applicant, main_employer_siret: entity.siret)}
    let!(:applicant2) {create(:applicant)}
    let!(:housing_request1) {create(:housing_request, is_active: true, documents_required: 50, main_applicant_id: applicant.id.to_s)}
    let!(:housing_request2) {create(:housing_request, is_active: true, documents_required: 50, main_applicant_id: applicant2.id.to_s)}
    
    before :each do
      entities_ids = entities.map {|e| e.id.to_s} + [entity.id.to_s]
      authenticated_header(request, superadministrateur)
      post 'search', {entity_id: {'$in': entities_ids}}
    end
    
    it 'returns status code 200' do
      expect(response).to have_http_status(200)
      expect(json['data'].length).to eq(1)
    end
    
  end
  
  describe 'GET /housing_requests' do
    let!(:housing_request1) {create(:housing_request, documents_required: 50, main_applicant_id: applicant.id.to_s)}
    let!(:housing_request3) {create(:housing_request, documents_required: 150, main_applicant_id: applicant.id.to_s)}
    let!(:housing_request2) {create(:housing_request, documents_required: 150, main_applicant_id: applicant.id.to_s)}
    let!(:nud1) {create(:nud, user_id: responsible.id.to_s, housing_request: housing_request1, unique_number: "0110418041931AL004")}
    let!(:nud2) {create(:nud, user_id: responsible.id.to_s, housing_request: housing_request2, unique_number: "0110418041931AL003")}
    let!(:nud3) {create(:nud, user_id: responsible.id.to_s, housing_request: housing_request3, unique_number: "0110418041931AL005")}
    
    [
      {user: 'responsible', total_objects: 3},
      {user: 'superadministrateur', total_objects: 3},
      {user: 'user', total_objects: 0},
    ].each do |a|
      describe "as #{a[:user]}" do
        before :each do
          authenticated_header(request, eval(a[:user]))
        end
        it 'returns a valid json' do
          get 'index'
          expect(response).to have_http_status(200)
          expect(json).not_to be_empty
          if (json["data"][0])
            expect(json['data'][0]["attributes"]["providers"]).to eq (["SNE"])
          end
          expect(json['data'].length).to eq(a[:total_objects])
          expect(json['meta']['pagination']).to be_present
          expect(json['meta']['pagination']['total_objects']).to eq(a[:total_objects])
        end

        it 'create log' do 
          get 'index', {main_applicant_id: applicant.id.to_s}
          Hermes::LogJob.perform_now
          log = Log.order_by(created_at: :desc).first
          expect(log.user_id.to_s).to eq(eval(a[:user]).id.to_s)
          expect(log.code).to eq("API-hermes/api/dmo/housing_requests")
          expect(log.action).to match(/status : 200 | index/)
          expect(log.data.json_to_h).to eq({main_applicant_id: applicant.id.to_s})
        end 
      end
    end
    describe "as admin" do
      before :each do
        
        authenticated_header(request, superadministrateur)
        get 'index', {unique_numbers: "0110418041931AL005"}
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
        expect(json['data'].length).to eq(1)
        expect(json['meta']['pagination']).to be_present
        expect(json['meta']['pagination']['total_objects']).to eq(1)
      end
    
    end
    describe 'GET with regex on first_name and last_name' do
      let!(:applicant2) {create(:applicant, first_name: 'Jean-Pierre', last_name: 'Dupié', auto_set_user: false)}
      let!(:applicant3) {create(:applicant, first_name: 'Jean', last_name: 'Dupont', auto_set_user: false)}
      let!(:housing_request4) {create(:housing_request, documents_required: 150, main_applicant_id: applicant2.id.to_s)}
      let!(:housing_request5) {create(:housing_request, documents_required: 150, main_applicant_id: applicant3.id.to_s)}
      
      before :each do
        authenticated_header(request, superadministrateur)
        get :index, {
          requester_firstname: {
            '$regex': 'jean-pierré'
          }
        }
      end
      
      it 'should return only one request' do
        expect(json['data'].count).to eq(1)
        expect(json['data'].first['id']).to eq(housing_request4.id.to_s)
      end
    end
    describe 'GET with regex on first_name and last_name' do
      let!(:applicant2) {create(:applicant, first_name: 'Jean-Pierre', last_name: 'Dupié', auto_set_user: false)}
      let!(:applicant3) {create(:applicant, first_name: 'Jean', last_name: 'Dupont', auto_set_user: false)}
      let!(:housing_request4) {create(:housing_request, documents_required: 150, main_applicant_id: applicant2.id.to_s)}
      let!(:housing_request5) {create(:housing_request, documents_required: 150, main_applicant_id: applicant3.id.to_s)}
      
      before :each do
        authenticated_header(request, superadministrateur)
        get :index, {
          applicant_used_for_siret_firstname: {
            '$regex': 'jean-pierré'
          }
        }
      end
      
      it 'should return only one request' do
        expect(housing_request4.applicant_used_for_siret_firstname).to eq('Jean-Pierre')
        expect(json['data'].count).to eq(1)
        expect(json['data'].first['id']).to eq(housing_request4.id.to_s)
      end
    end

    describe "send_flux_30" do
      let!(:applicant2) {create(:applicant, first_name: 'Jean-Pierre', last_name: 'Dupié', auto_set_user: false)}
      let!(:applicant3) {create(:applicant, first_name: 'Jean', last_name: 'Dupont', auto_set_user: false)}
      let!(:housing_request4) {create(:housing_request, documents_required: 150, main_applicant_id: applicant2.id.to_s)}
      let!(:housing_request5) {create(:housing_request, documents_required: 150, main_applicant_id: applicant3.id.to_s)}
      before :each do
        authenticated_header(request, superadministrateur)
        post :send_flux_thirty, {
          ids: [housing_request4.id.to_s, housing_request5.id.to_s]
        }
      end
      it 'should return only ok' do
        
        expect(json).to eq ({
          "res" => "2 housing_requests flux 30 sended to the workers"
        })
      
      end
    end
    describe "send_flux_30" do
      let!(:applicant2) {create(:applicant, first_name: 'Jean-Pierre', last_name: 'Dupié', auto_set_user: false)}
      let!(:applicant3) {create(:applicant, first_name: 'Jean', last_name: 'Dupont', auto_set_user: false)}
      let!(:housing_request4) {create(:housing_request, documents_required: 150, main_applicant_id: applicant2.id.to_s)}
      let!(:housing_request5) {create(:housing_request, documents_required: 150, main_applicant_id: applicant3.id.to_s)}
      before :each do
        authenticated_header(request, superadministrateur)
        drain_all_jobs
        post :send_flux_thirty, {
          als_ids: [housing_request4.reload.als_id.to_s, housing_request5.reload.als_id.to_s]
        }
      end
      it 'should return ok' do
        
        expect(json).to eq ({
          "res" => "2 housing_requests flux 30 sended to the workers"
        })
      
      end
    end

    describe 'GET available HousingRequest for current lessor' do
      let!(:lessor_user) {create(:user, first_name: 'Jaques', last_name: 'Dupont', email: 'jaques.dupont@gmail.com', roles: ['bailleur'])}
      let!(:lessor) {create(:lessor, user_id: lessor_user.id.to_s)}
      let!(:housing_offer) {create(:housing_offer, lessor: lessor)}
      let!(:housing_request) {create(:housing_request, main_applicant_id: applicant.id, is_active: true)}
      let!(:housing_application) {create(:housing_application, housing_request: housing_request, housing_offer: housing_offer)}
      before :each do
        authenticated_header(request, lessor_user)
        get :index
      end
      
      it 'should return only requests linked to current lessor offers' do
        expect(json['data'].count).to eq(1)
        expect(json['data'].first['id']).to eq(housing_request.id.to_s)
        expect(json['data'].first['attributes']['housing_application_ids']).to eq([housing_application.id.to_s])
      end
    end

    Hermes::Api::Dmo::HousingRequestsController.acceptable_params.each do |acceptable_param, options|
      next if ['main_applicant_id', 'viewed_by_colaborator_offre_logement', 'dependents', 'coapplicants', 'coapplicant_ids'].include?(acceptable_param.to_s)
      type = options[:type]
      describe "GET /housing_requests with sort on field #{acceptable_param} works" do
        let!(:housing_request3) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
        let!(:housing_request1) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
        let!(:housing_request2) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
        
        before :each do
          (1..3).each do |i|
            record = eval("housing_request#{i}")
            record.unset(acceptable_param.to_sym)
            record.set(acceptable_param.to_sym => get_typed_value(type, (i+1) % 3))
          end 

          authenticated_header(request, responsible)
          get 'index', {sort: {acceptable_param => 1}}
        end

        it 'returns sorted results' do
          attribute = if type == 'DateTime'
            json['data'].map {|unit| DateTime.parse(unit['attributes'][acceptable_param.to_s])}
          elsif type == 'Date'
            json['data'].map {|unit| Date.parse(unit['attributes'][acceptable_param.to_s])}
          else
            json['data'].map {|unit| unit['attributes'][acceptable_param.to_s]}
          end
          expect(response.status).to eq(200)
          expect(attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
        end
      end
    end
    
    describe 'GET /housing_requests with requester_email case insensitive' do
      let(:applicant1) {create :applicant, email: 'KevIN34515@Wanadoo.fr'}
      let(:applicant2) {create :applicant, email: 'brandon34@wanadoo.fr'}
      let!(:housing_request1) {create(:housing_request, main_applicant_id: applicant1.id)}
      let!(:housing_request3) {create(:housing_request, main_applicant_id: applicant2.id)}
      
      describe 'lower searching upper' do
      before do
        authenticated_header(request, responsible)
        get 'index', {requester_email: "kevin34515@wanadoo.fr"}
      end
      
      it 'returns results' do
        expect(response.status).to eq(200)
        expect(json['data'].length).to eq(1)
        expect(json['data'][0]['id']).to eq housing_request1.id.to_s
        
      end
    end
    describe 'upper searching lower' do
      before do
        authenticated_header(request, responsible)
        get 'index', {requester_email: "BRANDON34@WANADOO.FR"}
      end
      
      it 'returns results' do
        expect(response.status).to eq(200)
        expect(json['data'].length).to eq(1)
        expect(json['data'][0]['id']).to eq housing_request3.id.to_s
        
      end
    end
    end

    describe 'GET /housing_requests with main_employer_siret' do
      let(:applicant_siret1) {create :applicant, main_employer_siret: '12345600000000'}
      let(:applicant_siret2) {create :applicant, main_employer_siret: '98765400000000'}
      let!(:housing_request1) {create(:housing_request, main_applicant_id: applicant.id)}
      let!(:housing_request3) {create(:housing_request, main_applicant_id: applicant_siret1.id)}
      let!(:housing_request2) {create(:housing_request, main_applicant_id: applicant_siret2.id)}
      
      before do
        authenticated_header(request, responsible)
        get 'index', {main_employer_siret: ['12345600000000', '98765400000000']}
      end
      
      it 'returns results' do
        expect(response.status).to eq(200)
        expect(json['data'].length).to eq(2)
        expect(json['data'][0]['id']).to eq housing_request3.id.to_s
        expect(json['data'][1]['id']).to eq housing_request2.id.to_s
      end
    end
    
    describe 'GET /housing_requests with consent_level' do
      let!(:form_level_1) {create :entity_form_level, level: 2}
      let!(:form_level_2) {create :entity_form_level, level: 4}
      let!(:entity_form_1) {create :entity_form, housing_request_id: housing_request1.id, entity_form_level_id: form_level_1.id}
      let!(:entity_form_2) {create :entity_form, housing_request_id: housing_request2.id, entity_form_level_id: form_level_2.id}
      let!(:entity_form_3) {create :entity_form, housing_request_id: housing_request3.id, level: 2}
      let!(:housing_request4) {create :housing_request, main_applicant: applicant, cdii: true}
      let!(:entity_form_4) {create :entity_form, housing_request_id: housing_request4.id, level: 2}
      
      before do
        authenticated_header(request, responsible)
        get 'index', {consent_level: 2, cdii: {"$ne": true}}
      end
      
      it 'succeeds' do
        expect(response).to have_http_status 200
      end
      
      it 'returns results' do
        expect(json['data'].length).to eq 2
        ids = [housing_request1, housing_request3].map { |hr| hr.id.to_s }
        expect(json['data'].map {|hr| hr['id']}).to eq ids
      end
      
      context 'when entity_id' do
        let!(:entity) {create :entity, siret: '12345678901234'}
        let!(:app4) {create(:applicant, main_employer_siret: entity.siret)}
        let!(:housing_request4) {create :active_request, main_applicant: app4}
        let!(:entity_form_level) {create :entity_form_level, entity_id: entity.id, level: 4}
        let!(:entity_form_4) do
          create :entity_form, housing_request_id: housing_request4.id, entity_form_level_id: entity_form_level.id
        end
        
        before do
          get 'index', {consent_level: { '$in': [3, 4]}, entity_id: {"$in": [entity.id.to_s]}}
        end
        
        it 'succeeds' do
          expect(response).to have_http_status 200
        end
        
        it 'returns results' do
          expect(json['data'].length).to eq 1
          expect(json['data'].map {|hr| hr['id']}).to eq [housing_request4.id.to_s]
        end
      end
    end
  end
  
  describe "GET /housing_requests sort by cotations" do 
    let!(:hr1) { create(:active_request, cotations: {"33000" => 100}) }
    let!(:hr2) { create(:active_request, cotations: {"33000" => 300})}
    let!(:hr3) { create(:active_request, cotations: {"33000" => 200}) }
    let!(:hr4) { create(:active_request, cotations: {"75000" => 100}) }
    let!(:hr5) { create(:active_request, cotations: nil) }

    it 'works' do 
      authenticated_header(request, responsible)
      get 'index', sort: { 'cotations.33000': -1}

      expect(json['data'].length).to eq(5)
      expect(json['data'].map{ |d| d['attributes']['cotations'] && d['attributes']['cotations']['33000']}).to eq([300, 200, 100, nil, nil])
    end 
  end 

  describe "GET /housing_requests filter by target_offer_reserving_business" do 
    let!(:hr1) { create(:active_request) }
    let!(:hr2) { create(:active_request)}

    before do 
      hr1.set(target_offer_reserving_business: 'ALS')
    end 


    it 'works' do 
      authenticated_header(request, responsible)
      get 'index', target_offer_reserving_business: 'ALS'

      expect(json['data'].length).to eq(1)
      expect(json['data'].first['id']).to eq(hr1.id.to_s)
      expect(json['data'].first['attributes']['target_offer_reserving_business']).to eq('ALS')

    end 
  end 

  describe 'GET /housing_requests/:id' do
    describe 'verif it work' do
      let!(:housing_request1) {create(:housing_request, motivations: ['acceuillant'], documents_required: 50, main_applicant_id: applicant.id.to_s)}
      
      it_behaves_like "true user all allowed" do
        let(:path) {['show', {id: housing_request1.id}]}
      end
      
      before :each do
        authenticated_header(request, responsible)
        get 'show', {id: housing_request1.id}
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
      end
    end
    describe 'get active housing request for current_user' do
      
      let!(:demandeur) {create(:user, auth_roles: ['demandeur'], last_name: 'Doe', first_name: 'John')}
      let!(:demandeur_applicant) {create(:applicant, user: demandeur)}
      let!(:housing_request1) {create(:housing_request, is_active: true, main_user_id: demandeur.id, responsible: demandeur.id, motivations: ['acceuillant'], documents_required: 50, main_applicant_id: demandeur_applicant.id.to_s)}
      
      before :each do
        authenticated_header(request, demandeur)
        get 'show', {id: 'active'}
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
        expect(json["data"]["id"]).to eq(housing_request1.id.to_s)
        expect(json["data"]["attributes"]["motivations"]).to eq(['acceuillant'])
      end
    end
    describe 'check is nud is rendrered' do
      let!(:housing_request1) {create(:housing_request, documents_required: 50, main_applicant_id: applicant.id.to_s)}
      let!(:nud1) {create(:nud, user_id: responsible.id.to_s, housing_request: housing_request1, unique_number: "ZERFEF")}
      before :each do
        authenticated_header(request, responsible)
        get 'show', {id: housing_request1.id}
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
        expect(json["data"]["attributes"]["unique_numbers"]).to eq(["ZERFEF"])
      end
    end
    describe 'verif it work not work with a non authorized user' do
      let!(:housing_request1) {create(:housing_request,  documents_required: 50, main_applicant_id: applicant.id.to_s)}
      
      before :each do
        authenticated_header(request, user)
        get 'show', {id: housing_request1.id}
      end
      it 'returns status code 404' do
        expect(response).to have_http_status(404)
      end
    end
    describe 'eligible_count option' do
      let!(:hr) {create(:complete_request, seeked_district_insee_codes: ['34000'], als_id: '0000002')}
      let!(:offer) { create(:published_offer, typology: 'T1', ceil_category_1: 25000, floor_category_1: 0, insee_code: '34000') }

      it 'not used by default' do 
        authenticated_header(request, superadministrateur)
        get 'show', {id: hr.als_id}
        expect(response).to have_http_status(200)
        expect(json["data"]["attributes"]["eligible_count"]).to eq(nil)
      end 

      it 'works' do 
        authenticated_header(request, superadministrateur)
        get 'show', {id: hr.id, options: [:eligible_count]}
        expect(response).to have_http_status(200)
        expect(json["data"]["attributes"]["eligible_count"]["department_count"]).to eq(1)
        expect(json["data"]["attributes"]["eligible_count"]["seeked_count"]).to eq(1)
        expect(json["data"]["attributes"]["eligible_count"]["bordering_count"]).to eq(0)
        expect(json["data"]["attributes"]["eligible_count"]["seeked_bordering_count"]).to eq(1)
        expect(json["data"]["attributes"]["eligible_count"]["other_from_department_count"]).to eq(0)
        expect(json["data"]["attributes"]["eligible_count"]["total_count"]).to eq(1)
      end

      it 'error when not eligible' do
        hr.set(eligibility_query_base: nil)
        authenticated_header(request, superadministrateur)
        get 'show', {id: hr.id, options: [:eligible_count]}
        expect(response).to have_http_status(422)
      end 

    end 
  end

  describe 'GET /housing_requests/:id/eligible_offers' do
    let!(:app) {create(:applicant, :with_eligible_offers, offers_count: 3)}
    let!(:hr) { app.first_active_request}
  
    before :each do 
      hr.set(als_id: '0000001')
    end 

    it 'error when eligiblity missing' do
      hr.set(eligibility_query_base: nil) 
      authenticated_header(request, responsible)
      get 'eligible_offers', id: hr.als_id
      expect(response).to have_http_status(422)
    end 
    
    it 'works' do 
      authenticated_header(request, responsible)
      get 'eligible_offers', id: hr.als_id
      expect(json['data'].length).to eq(3)
    end 

    it 'works with filter' do
      hr.eligibles_offers.order_by(created_at: :desc).first.set(typology: 'T3') 
      hr.eligibles_offers.order_by(created_at: :desc).last.set(typology: 'T4') 
      authenticated_header(request, responsible)
      get 'eligible_offers', id: hr.als_id
      expect(json['data'].length).to eq(2)
      get 'eligible_offers', id: hr.als_id, typology: 'T3'
      expect(json['data'].length).to eq(1)
      get 'eligible_offers', id: hr.als_id, typology: 'T4'
      expect(json['data'].length).to eq(0)
    end 
    
    it 'works with type and count' do 
      hr.eligibles_offers.first.set(insee_code: '34999')
      authenticated_header(request, responsible)
      get 'eligible_offers', id: hr.als_id, eligibility_type: 'seeked', options: ['other_from_department_count', 'bordering_count']
      expect(json['data'].length).to eq(2)
      expect(json['meta']['other_from_department_count']).to eq(1)
      expect(json['meta']['bordering_count']).to eq(0)
    end 
  end 

  describe 'GET /housing_requests/:id/is_eligible_offer' do
    let!(:app) {create(:applicant, :with_eligible_offers, offers_count: 1)}
    let!(:hr) { app.first_active_request}
    let!(:eligible_offer) { hr.eligibles_offers.first}
    let!(:other_offer) { create(:housing_offer)}

    before :each do 
      authenticated_header(request, responsible)
    end 

    it 'not found instance' do 
      get 'is_eligible_offer', id: 'not found', offer_id: eligible_offer.id.to_s
      expect(response).to have_http_status(404)
      expect(json['errors'][0]['title']).to include('instance_not_found')
    end 

    it 'missing offer_id' do 
      get 'is_eligible_offer', id: hr.id.to_s, offer_id: ''
      expect(response).to have_http_status(406)
      expect(json['errors'][0]['title']).to include('missing_offer_id')
    end 

    it 'not found offer' do 
      get 'is_eligible_offer', id: hr.id.to_s, offer_id: 'not found'
      expect(response).to have_http_status(404)
      expect(json['errors'][0]['title']).to include('offer_not_found')
    end 

    it 'works' do 
      get 'is_eligible_offer', id: hr.id.to_s, offer_id: eligible_offer.id.to_s
      expect(response).to have_http_status(200)
      expect(json['is_eligible']).to eq(true)
    end 

    it 'works' do 
      get 'is_eligible_offer', id: hr.id.to_s, offer_id: other_offer.id.to_s
      expect(response).to have_http_status(200)
      expect(json['is_eligible']).to eq(false)
    end 

  end 
  
  describe 'patch request with entity_validated is valid' do
    let!(:housing_request) {create(:housing_request, user: user2, main_applicant_id: applicant.id.to_s, is_active: true)}
    let!(:user2) {create(:user, auth_roles: ["Entreprise"])}
    let!(:entity) {create(:entity, user: user2, siret: '12345678901234')}
    
    before :each do
      applicant.update(main_employer_siret: entity.siret)
      authenticated_header(request, user2)
      patch 'update', {
        id: housing_request.id.to_s,
        data: {
          attributes: {entity_validated: true}
        }
      }
    end
    it 'returns status code 200 and the value appears in json' do
      expect(response).to have_http_status(200)
      expect(housing_request.reload.entity_validated).to eq(true)
      expect(housing_request.reload.last_update_user_id).to eq(user2.id.to_s)
    end
  end
  
  describe 'patch request with entity_validated is not valid when done by user' do
    let!(:applicant3) {create(:applicant, user_id: user.id)}
    let!(:housing_request) {create(:housing_request, main_applicant_id: applicant3.id.to_s, user_id: user.id, is_active: true)}
    let!(:user2) {create(:user, auth_roles: ["Entreprise"])}
    let!(:entity) {create(:entity, user_id: user2.id, siret: '12345678901234')}
    
    before :each do
      applicant.main_employer_siret = entity.siret
      applicant.save
      authenticated_header(request, user)
      patch 'update', {
        id: housing_request.id.to_s,
        data: {
          attributes: {entity_validated: true}
        }
      }
    end
    it 'returns status code 200 and the value appears in json' do
      expect(response).to have_http_status(422)
      drain_all_jobs
      expect(housing_request.reload.entity_validated).to eq(false)
    end
  end

  describe 'DELETE /housing_requests/:id' do
    let!(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s, user_id: responsible.id.to_s)}
    before :each do
      authenticated_header(request, responsible)
      
      delete 'destroy', {
        id: housing_request.id.to_s
      }
    end
    it 'returns status code 404' do
      expect(response).to have_http_status(204)
    end
  end
  
  describe 'POST /housing_requests/:id/validate_code' do
    let!(:entity1) {create :entity, siret: '00000000000001'}
    let!(:entity2) {create :entity, siret: '00000000000002'}
    let!(:expired_code) {create :entity_code}
    let!(:applicant) {create :applicant}
    let!(:entity_code_with_applicant) {create :entity_code, applicant_id: applicant.id, entity_id: entity1.id, code: 'ABC123'}
    let!(:entity_code_without_applicant) {create :entity_code, entity_id: entity2.id, code: 'DEF456'}
    let!(:housing_request) {create :housing_request, main_applicant_id: applicant.id, is_active: true}
    let!(:entity_form) {create :entity_form, housing_request_id: housing_request.id, level: 3}
    
    before do
      authenticated_header(request, responsible)
    end
    
    it 'requires code parameter' do
      post 'validate_code', {id: housing_request.id.to_s}
      expect(response).to have_http_status 406
    end
    
    it 'returns not_found error' do
      post 'validate_code', {id: housing_request.id.to_s, code: 'unknown_code'}
      expect(response).to have_http_status 404
      expect(json['success']).to eq false
      expect(json['message']).to eq 'Le code n\'est pas valide.'
    end
    
    context 'when entity_code is expired' do
      before do
        expired_code.set(expire_at: Time.zone.now.yesterday)
        post 'validate_code', {id: housing_request.id.to_s, code: expired_code.code}
      end
      it 'returns expired code error' do
        expect(response).to have_http_status 422
        expect(json['success']).to eq false
        expect(json['message']).to eq 'Le code est expiré.'
      end
    end
    
    context 'when entity_code_with_applicant' do
      it 'succeeds' do
        post 'validate_code', {id: housing_request.id.to_s, code: 'ABC123'}
        expect(response).to have_http_status 200
      end
      it 'sets the last_entity_code' do
        expect do
          post 'validate_code', {id: housing_request.id.to_s, code: entity_code_with_applicant.code}
        end.to change {housing_request.reload.last_entity_code}.from(nil).to 'ABC123'
      end
    end
    
    context 'when entity_code_without_applicant' do
      it 'succeeds' do
        post 'validate_code', {id: housing_request.id.to_s, code: 'DEF456'}
        expect(response).to have_http_status 200
      end
      it 'sets the last_entity_code' do
        expect do
          post 'validate_code', {id: housing_request.id.to_s, code: 'DEF456'}
        end.to change {housing_request.reload.last_entity_code}.from(nil).to 'DEF456'
      end
    end

    context 'when code is empty' do
      it 'succeeds' do
        applicant.update(main_employer_siret: entity2.siret)
        expect(housing_request.reload.entity_id).to eq(entity2.id)
        post 'validate_code', {id: housing_request.id.to_s, code: entity_code_with_applicant.code}
        expect(housing_request.reload.entity_id).to eq(entity1.id)
        post 'validate_code', {id: housing_request.id.to_s, code: nil}
        expect(response).to have_http_status 200
        expect(housing_request.reload.entity_id).to eq(entity2.id)
      end
    end

    it 'reload when entity_id is nil' do 
      applicant.update(main_employer_siret: entity1.siret)
      housing_request.update(last_entity_code: entity_code_with_applicant.code)
      housing_request.reload
      applicant.reload
      expect(housing_request.entity_id).to eq(entity1.id)
      expect(applicant.main_employer_siret).to eq(entity1.siret)
      applicant.update(main_employer_siret: entity2.siret)
      housing_request.reload
      applicant.reload
      expect(housing_request.entity_id).to eq(entity1.id)
      expect(applicant.main_employer_siret).to eq(entity2.siret)
      post 'validate_code', {id: housing_request.id.to_s, code: nil}
      attributes = json['data']['attributes']
      housing_request.reload
      applicant.reload
      expect(housing_request.entity_id).to eq(entity2.id)
      expect(attributes['entity_id']).to eq(entity2.id.to_s)
    end 

    it 'clear consent level' do 
      expect(Hermes::HousingRequestEntityForm.where(housing_request_id: housing_request.id).count).to eq(1)
      expect do
        post 'validate_code', {id: housing_request.id.to_s, code: 'DEF456'}
      end.to change {housing_request.reload.consent_level}.from(3).to(nil)
      expect(Hermes::HousingRequestEntityForm.where(housing_request_id: housing_request.id).count).to eq(0)
    end 

  end
  
  describe "housing channels management" do
    let!(:channel_1) {create :housing_channel, key: 'dalo'}
    let!(:channel_2) {create :housing_channel, key: 'priority_public'}
    let!(:channel_3) {create :housing_channel, key: 'droits_communs'}
    let!(:channel_4) {create :housing_channel, key: 'key1'}
    let!(:channel_5) {create :housing_channel, key: 'key2'}
    let!(:channel_setting) {create :housing_channel_setting, department: '78', housing_channel_keys: ['dalo', 'priority_public', 'key1', 'key2']}
    
    let(:housing_request) do
      create(
        :housing_request,
        main_applicant: applicant,
        seeked_district_postal_codes: ['78000', '75000'],
        housing_channel_pending_keys: ['dalo'],
        housing_channel_chosen_keys: ['dalo'],
        housing_channel_keys: ['priority_public']
      )
    end
    before do
      authenticated_header(request, responsible)
    end
    
    describe 'POST /housing_requests/:id/add_housing_channel' do
      it 'requires key parameter' do
        post 'add_housing_channel', {id: housing_request.id.to_s}
        expect(response).to have_http_status 406
        expect(json['errors'][0]['message']).to eq 'param is missing or the value is empty: key or keys'
      end
      
      it 'returns not found channel error' do
        post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['unknown']}
        expect(response).to have_http_status 404
        expect(json['errors'][0]['message']).to eq "La filière logement 'unknown' n'existe pas."
      end
      
      it 'returns already added error' do
        post 'add_housing_channel', {id: housing_request.id.to_s, key: 'priority_public'}
        expect(response).to have_http_status 409
        expect(json['errors'][0]['message']).to eq "La filière logement 'priority_public' a déjà été ajoutée."
      end
      
      it 'returns unavailable for department error' do
        post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['droits_communs']}
        expect(response).to have_http_status 422
        expect(json['errors'][0]['message']).to eq "La filière logement 'droits_communs' n'est pas disponible pour les départements '75, 78' recherchés."
      end
      
      xit 'succeeds' do
        expect do
          post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['priority_public']}
        end.to change {housing_request.reload.housing_channel_pending_keys}.from(['dalo']).to(['dalo', 'priority_public'])
                  .and change {housing_request.reload.housing_channel_keys}.from(['priority_public', 'dc']).to ['dc']
      end

      it 'succeeds with keys' do
        expect do
          post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['key1', 'key2']}
        end.to change {housing_request.reload.housing_channel_keys}.from(['priority_public', 'dc']).to(['key1', 'dc', 'key2'])
        expect(response).to have_http_status 200
      end

      it 'succeeds with key' do
        expect do
          post 'add_housing_channel', {id: housing_request.id.to_s, key: 'key1'}
        end.to change {housing_request.reload.housing_channel_keys}.from(['priority_public', 'dc']).to(['priority_public', 'dc', 'key1'])
        expect(response).to have_http_status 200
      end
      
      xit 'sets to false when boolean field exists' do
        expect do
          post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['priority_public']}
        end.to change {housing_request.reload.priority_public}.to false
      end
      
      xit 'adds the key to the housing_channel_chosen_keys' do
        expect do
          post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['priority_public']}
        end.to change {housing_request.reload.housing_channel_chosen_keys}.from(['dalo', 'dc']).to(['dalo', 'dc', 'priority_public'])
      end

      it 'can add double channel' do
        expect do
          post 'add_housing_channel', { id: housing_request.id.to_s, keys: ['dalo',  'dalo']}
        end.to change { housing_request.reload.housing_channel_keys }.from(['priority_public', 'dc']).to(['dalo', 'dc'])
        expect(response).to have_http_status 200
      end
      
      #===== TEMPORARY BYPASS VALIDATION =======
      context 'with the temporary bypass' do
        let(:housing_request) do
          create(
            :housing_request,
            main_applicant: applicant,
            seeked_district_postal_codes: ['78000', '75000'],
            housing_channel_chosen_keys: ['dalo'],
            housing_channel_keys: ['dalo']
          )
        end
        
        it 'succeeds' do
          expect do
            post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['priority_public']}
          end.to change {housing_request.reload.housing_channel_keys}.from(['dalo', 'dc']).to ['priority_public', 'dc']
        end
        it 'succeeds to remove' do
          expect do
            post 'remove_housing_channel', {id: housing_request.id.to_s, keys: ['dalo']}
          end.to change {housing_request.reload.housing_channel_keys}.from(['dalo', 'dc']).to ['dc']
        end
        it 'sets to true when boolean field exists' do
          expect do
            post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['priority_public']}
          end.to change {housing_request.reload.priority_public}.to true
        end
        
        it 'adds the key to the housing_channel_chosen_keys' do
          expect do
            post 'add_housing_channel', {id: housing_request.id.to_s, keys: ['priority_public']}
          end.to change {housing_request.reload.housing_channel_chosen_keys}.from(['dalo']).to(['dalo', 'priority_public'])
        end
      end
      #=========================================
    end
    
    describe 'POST /housing_requests/:id/validate_housing_channel' do
      it 'requires key parameter' do
        post 'validate_housing_channel', {id: housing_request.id.to_s}
        expect(response).to have_http_status 406
        expect(json['errors'][0]['message']).to eq 'param is missing or the value is empty: key or keys'
      end
      
      it 'returns not found channel error' do
        post 'validate_housing_channel', {id: housing_request.id.to_s, keys: ['unknown']}
        expect(response).to have_http_status 404
        expect(json['errors'][0]['message']).to eq "La filière logement 'unknown' n'existe pas."
      end
      
      it 'returns already validated error' do
        post 'validate_housing_channel', {id: housing_request.id.to_s, keys: ['priority_public']}
        expect(response).to have_http_status 422
        expect(json['errors'][0]['message']).to eq "La filière logement 'priority_public' a déjà été validée."
      end
      
      it 'returns not in pending list error' do
        post 'validate_housing_channel', {id: housing_request.id.to_s, keys: ['droits_communs']}
        expect(response).to have_http_status 422
        expect(json['errors'][0]['message']).to eq "La filière logement 'droits_communs' n'est pas dans la liste à valider."
      end
      
      it 'succeeds' do
        expect do
          post 'validate_housing_channel', {id: housing_request.id.to_s, keys: ['dalo']}
        end.to change {housing_request.reload.housing_channel_pending_keys}.from(['dalo']).to([])
                  .and change {housing_request.reload.housing_channel_keys}.from(['priority_public', 'dc']).to ['priority_public', 'dc', 'dalo']
      end
      
      it 'sets to true when boolean field exists' do
        expect do
          post 'validate_housing_channel', {id: housing_request.id.to_s, keys: ['dalo']}
        end.to change {housing_request.reload.dalo}.to true
      end
      
      it 'does not remove the key from housing_channel_chosen_keys' do
        expect(housing_request.housing_channel_pending_keys).to eq ['dalo']
        expect(housing_request.housing_channel_chosen_keys).to eq ['dalo']
        post 'validate_housing_channel', {id: housing_request.id.to_s, keys: ['dalo']}
        
        expect(housing_request.reload.housing_channel_pending_keys).to eq []
        expect(housing_request.reload.housing_channel_chosen_keys).to eq ['dalo']
      end
      
      context 'on sending flux 35' do
        before do
          VCR.use_cassette("request_orientation_task1", :match_requests_on => [:method, :uri, :path]) do
            post 'validate_housing_channel', {id: housing_request.id.to_s, keys: ['dalo']}
          end
        end
        it 'is OK' do
        end
      end
    end
  end
  
  # HOUSING REQUEST FOLDER
  describe 'GET /housing_requests/:id/housing_request_folder' do
    let!(:user1) {create(:user)}
    let!(:user2) {create(:user)}
    let!(:applicant42) {
      create(
        :applicant,
        user: user1,
        monthly_charges: {},
        nationality: "3"
      )
    }
    let!(:applicant2) {
      create(
        :applicant,
        user: user2,
        monthly_charges: {},
        nationality: "2"
      )
    }
    let!(:housing_request5) {
      create(
        :housing_request,
        :with_dependents,
        is_active: true,
        main_applicant: applicant42,
        coapplicants: [{ rel_type: 'P', id:applicant2.id.to_s }],
      )
    }
    describe 'get ok' do
      before :each do
        Hermes::RequiredDocument.all.each do |required_doc|
          required_doc.declared = true
          required_doc.save
        end
        authenticated_header(request, responsible)
        get 'housing_request_folder', {
          id: housing_request5.id.to_s
        }
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
        expect(json["data"]["attributes"]["required_documents"].length).to eq(7)
        expect(json["data"]["attributes"]["nud_status"]).to eq("Dossier validé")
      end
    end
    describe 'get ok' do
      before :each do
        file = File.open([Hermes::Engine.root, %w[spec xml_bouchon minion.jpg]].flatten.join('/'))
        blob = file.read
        Hermes::RequiredDocument.all.each do |required_doc|
          
          required_doc.save
          required_doc.declared = true
          required_doc.save
        end
        authenticated_header(request, responsible)
        get 'housing_request_folder', {
          id: housing_request5.id.to_s
        }
      end
      it 'returns a valid json' do
        expect(response).to have_http_status(200)
        expect(json).not_to be_empty
        expect(json["data"]["attributes"]["required_documents"].length).to eq(7)
        expect(json["data"]["attributes"]["required_documents"][0]["status"]).to eq("Pièce validée")
        expect(json["data"]["attributes"]["nud_status"]).to eq("Dossier validé")
      end
    end
  

    describe 'change_confirmed' do 
      before :each do
        file = File.open([Hermes::Engine.root, %w[spec xml_bouchon minion.jpg]].flatten.join('/'))
        blob = file.read
        
        Hermes::RequiredDocument.all.each do |required_doc|
        
          required_doc.save
        end
        authenticated_header(request, responsible)
      end

      it 'error when missing params' do 
        post 'change_confirmed', {
          id: housing_request5.id.to_s
        }
        expect(response).to have_http_status(422)
        expect(housing_request5.housing_request_folder.required_documents.map { |rd| rd.reload.status}.uniq).not_to eq(["Pièce validée"])
      end 

      it 'when all' do 
        post 'change_confirmed', {
          id: housing_request5.id.to_s,
          required_documents:{
            all: true
          }
        }
        expect(response).to have_http_status(200)
        expect(housing_request5.housing_request_folder.required_documents.map { |rd| rd.reload.status}.uniq).to eq(["Pièce validée"])
      end 

      it 'when ids' do 
        hrf = housing_request5.housing_request_folder
        rd1 = hrf.required_documents[0]
        rd2 = hrf.required_documents[1]
        post 'change_confirmed', {
          id: housing_request5.id.to_s,
          required_documents:{
            rd1.id => true,
            rd2.id => false,
          }
        }
        expect(response).to have_http_status(200)
        expect(rd1.reload.status).to eq("Pièce validée")
        expect(rd2.reload.status).to eq("Pièce manquante")
      end 

      
    end 
  end
  
  describe 'GET /housing_requests/:id/housing_request_folder auto replace olds names' do
    let!(:user1) {create(:user)}
    let!(:user2) {create(:user)}
    let!(:applicant42) {
      create(
        :applicant,
        user: user1,
        monthly_charges: {},
        nationality: "3"
      )
    }
    
    let!(:housing_request5) {
      create(
        :housing_request,
        user: user1,
        is_active: true,
        main_applicant: applicant42,
      )
    } 

    describe 'reset_pj' do
      before :each do
        Hermes::RequiredDocument.all.each do |required_doc|
          required_doc.declared = true
          required_doc.save
        end
        authenticated_header(request, responsible)
        patch 'update', {
          id: housing_request5.id.to_s,
          data: {
            attributes: {
              is_active: false,
              form_status: "cancelled"
            }
          }
        }
        patch 'update', {
          id: housing_request5.id.to_s,
          data: {
            attributes: {
              is_active: true
            }
          }
        }
      end
      it 'returns results' do
        housing_request5.reload.housing_request_folder.required_documents.each do |rd|
          expect(rd.declared).to eq(false)
          expect(rd.status).to eq('Pièce manquante')
      
        end
        
      end
    end
    describe 'get ok' do
      before :each do
        rd = housing_request5.housing_request_folder.required_documents[1]
        rd.type = "Avis d'imposition N-2"
        rd.save
    
        authenticated_header(request, responsible)
        get 'housing_request_folder', {
          id: housing_request5.id.to_s
        }
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'object is renamed in db if folder recompute' do
        housing_request5.motivations = ["08"]
        housing_request5.save
        rd = housing_request5.reload.housing_request_folder.reload.required_documents[1]
        expect(rd.reload.type).to eq("Avis d'imposition "+ ((Date.current - 1.years).year.to_s )+" sur les revenus de l'année " + ((Date.current - 2.years).year.to_s ))
      
      end
    end
    
  end
  
  describe 'desactivate and delete' do
    let!(:admin) {create(:user, roles: ['Administrator'])}
    let!(:user_applicant) {create(:user, auth_roles: ['demandeur'])}
    let!(:main_applicant) {create(:applicant, user: user_applicant)}
    let!(:housing_request1) {create(:housing_request, user: admin, main_applicant_id: main_applicant.id, is_active: true)}
    let!(:housing_request2) {create(:housing_request, user: admin, main_applicant_id: main_applicant.id, is_active: true)}
    let!(:housing_request3) {create(:housing_request, user: user_applicant, main_applicant_id: main_applicant.id)}
    let!(:housing_offer1) {create(:housing_offer)}
    let!(:housing_offer2) {create(:housing_offer)}
    let!(:housing_offer3) {create(:housing_offer)}
    let!(:housing_application1) {create(:housing_application, housing_request: housing_request1, housing_offer: housing_offer1)}
    let!(:housing_application2) {create(:housing_application, housing_request: housing_request1, housing_offer: housing_offer2)}
    let!(:housing_application3) {create(:housing_application, housing_request: housing_request3, housing_offer: housing_offer2)}
    
    describe 'PATCH /housing_requests/:id' do
      describe 'on request with applications' do
        before :each do
          authenticated_header(request, admin)
          patch 'update', {
            id: housing_request1.id.to_s,
            data: {
              attributes: {
                is_active: false
              }
            }
          }
        end
        it 'returns results' do
          expect(housing_request1.housing_offer_ids).to eq([housing_offer1.id.to_s, housing_offer2.id.to_s])
          expect(response.status).to eq(422)
        end
      end
      
      
      describe 'on request without applications' do
        before :each do
          authenticated_header(request, admin)
          patch 'update', {
            id: housing_request2.id.to_s,
            data: {
              attributes: {
                is_active: false
              }
            }
          }
        end
        it 'returns results' do
          expect(response.status).to eq(200)
        end
      end
    end
    
    describe 'DELETE /housing_requests/:id' do
      describe 'on request with applications' do
        before :each do
          authenticated_header(request, admin)
          delete 'destroy', {
            id: housing_request1.id.to_s
          }
        end
        it 'returns results' do
          expect(response.status).to eq(422)
        end
      end
      describe 'on request without applications' do
        before :each do
          authenticated_header(request, admin)
          delete 'destroy', {
            id: housing_request2.id.to_s
          }
        end
        it 'returns results' do
          expect(response.status).to eq(204)
        end
      end
    end
    
    describe 'unable_to_update_if application' do
      before :each do
        authenticated_header(request, user_applicant)
        patch 'update', {
          id: housing_request3.id.to_s,
          data: {
            attributes: {
              with_car_park: true
            }
          }
        }
      end
      xit 'returns results' do
        expect(response.status).to eq(422)
      end
    end
    describe 'allowed_for_admin' do
      before :each do
        authenticated_header(request, admin)
        patch 'update', {
          id: housing_request3.id.to_s,
          data: {
            attributes: {
              with_car_park: true
            }
          }
        }
      end
      it 'returns results' do
        expect(response.status).to eq(200)
      end
    end
  end
  
  context 'Superadmin searches' do
    let!(:superadmin) {create(:user, auth_roles: ['superadministrateur'])}
    let!(:user) {create(:user)}
    let!(:main_applicant) {
      create(
        :applicant,
        user: superadmin,
        monthly_charges: {},
        nationality: "3"
      )
    }
    
    let!(:applicant_id) {
      create(
        :applicant,
        user: user,
        monthly_charges: {},
        nationality: "2"
      )
    }
    let!(:entity) {create(:entity, title: "test", siret: "12345671234567", id_cristal: "1")}
    
    let!(:housing_request) {
      create(
        :housing_request,
        main_applicant: main_applicant,
        applicant_ids: [applicant_id.id.to_s])
      
    }
  end
  
  context 'Superadmin searches' do
    let!(:superadmin) {create(:user, auth_roles: ['superadministrateur'], first_name: "test_first", last_name: "test_last")}
    let!(:entity1) {create(:entity, title: "test", siret: "00000000000001", id_cristal: "1")}
    let!(:entity2) {create(:entity, title: "test2", siret: "00000000000002", id_cristal: "2")}
    let!(:entity3) {create(:entity, title: "test3", siret: "00000000000003", id_cristal: "3")}
    let!(:applicant1) {create(:applicant, first_name: "John", last_name: "Doe", main_employer_siret: entity1.siret)}
    let!(:applicant2) {create(:applicant, first_name: "Jane", last_name: "Doe", main_employer_siret: entity1.siret)}
    let!(:applicant3) {create(:applicant, first_name: "Bob", last_name: "Dupont", main_employer_siret: entity2.siret)}
    let!(:houing_request1) {create(:housing_request, main_applicant_id: applicant1.id, is_active: true)}
    let!(:houing_request2) {create(:housing_request, main_applicant_id: applicant2.id, is_active: true)}
    let!(:houing_request3) {create(:housing_request, main_applicant_id: applicant3.id, is_active: true)}
    let!(:nud) {create(:nud, applicant_id: applicant2.id, housing_request_id: houing_request2.id )}
    
    describe 'GET /housing_request/admin_search' do
      describe "search with user that is not a Superadmin" do
        before do
          authenticated_header(request, user)
          get 'admin_search'
        end
        
        it "refuses the access" do
          body = JSON.parse(response.body)
          expect(body["code"]).to eq(401)
        end
      end
      
      describe "search with good params" do
        
        before do
          authenticated_header(request, superadmin)
          get 'admin_search', {
            entity_params: {siret: entity1.siret}
          }
        end
        
        it "as valid data" do
          datas = json['data']
          expect(datas.length).to eq(2)
          expect(datas[0]['entity_title']).to eq(entity1.title)
          expect(datas[0]['applicant_first_name']).to eq(applicant1.first_name)
        end
      end
      
      describe "search with good params + user params" do
        
        before do
          authenticated_header(request, superadmin)
          get 'admin_search', {
            entity_params: {siren: entity1.siren},
            applicant_params: {requester_firstname: applicant2.first_name}
          }
        end
        
        it "as valid data" do
          datas = json['data']
          expect(datas.length).to eq(1)
          expect(datas[0]['id']).to eq(houing_request2.id.to_s)
          expect(datas[0]['entity_title']).to eq(entity1.title)
          expect(datas[0]['applicant_first_name']).to eq(applicant2.first_name)
        end
      end
      
      describe "search with nud params" do
        
        before do
          authenticated_header(request, superadmin)
          get 'admin_search', {
            unique_numbers: [nud.unique_number]
          }
        end
        
        it "as valid data" do
          datas = json['data']
          expect(datas.length).to eq(1)
          expect(datas[0]['id']).to eq(houing_request2.id.to_s)
          expect(datas[0]['unique_numbers']).to eq([nud.unique_number])
          expect(datas[0]['applicant_first_name']).to eq(applicant2.first_name)
        end
      end
    end
  end
  
  context "entity notes and internal notes" do
    describe "GET /housing_requests/:id/entity_note" do
      let!(:user1) {create(:user, auth_roles: ['cre'])}
      let!(:user2) {create(:user)}
      let!(:entity1) {create(:entity, siret: "00000000000001", writers: [user1.id.to_s])}
      let!(:applicant42) {
        create(
          :applicant,
          user: user1,
          monthly_charges: {},
          nationality: "3",
          main_employer_siret: entity1.siret,
        )
      }
      
      let!(:applicant2) {
        create(
          :applicant,
          user: user2,
          monthly_charges: {},
          nationality: "2"
        )
      }
      
      let!(:housing_request5) {
        create(
          :housing_request,
          :with_dependents,
          is_active: true,
          main_applicant: applicant42,
          applicant_ids: [applicant2.id.to_s],
          entity_note: "test note",
          writers: [user1.id.to_s]
        )
      }
      before do
        authenticated_header(request, user1)
        get 'entity_note', {
          id: housing_request5.id.to_s
        }
      end
      
      it 'returns the proper entity note' do
        resp = JSON.parse(response.body)
        expect(resp['message']).to eq(housing_request5.entity_note)
      end
    end
    
    describe "GET /housing_requests/:id/internal_note" do
      let!(:user_collab) {create(:user, auth_roles: ['collaborateur offre logement'])}
      let!(:user2) {create(:user)}
      let!(:entity1) {create(:entity, siret: "00000000000001")}
      let!(:applicant42) {
        create(
          :applicant,
          user: user_collab,
          monthly_charges: {},
          nationality: "3",
          main_employer_siret: entity1.siret,
        )
      }
      
      let!(:applicant2) {
        create(
          :applicant,
          user: user2,
          monthly_charges: {},
          nationality: "2"
        )
      }
      
      let!(:housing_request5) {
        create(
          :housing_request,
          :with_dependents,
          is_active: true,
          main_applicant: applicant42,
          applicant_ids: [applicant2.id.to_s],
          internal_note: "test note"
        )
      }
      before do
        authenticated_header(request, user_collab)
        get 'internal_note', {
          id: housing_request5.id.to_s
        }
      end
      
      it 'returns the proper internal note' do
        resp = JSON.parse(response.body)
        expect(resp['message']).to eq(housing_request5.internal_note)
      end
    end
    
    describe "POST /housing_requests/:id/update_internal_note" do
      let!(:user_collab) {create(:user, auth_roles: ['collaborateur offre logement'])}
      let!(:user2) {create(:user)}
      let!(:applicant_collab) {
        create(
          :applicant,
          user: user_collab,
          monthly_charges: {},
          nationality: "3"
        )
      }
      let!(:applicant2) {
        create(
          :applicant,
          user: user2,
          monthly_charges: {},
          nationality: "2",
        )
      }
      let!(:entity1) {create(:entity)}
      
      let!(:housing_request_true) {
        create(
          :housing_request,
          :with_dependents,
          is_active: true,
          main_applicant: applicant_collab,
          applicant_ids: [applicant2.id.to_s],
          internal_note: "test note"
        )
      }
      before do
        authenticated_header(request, user_collab)
      end
      
      it 'returns the proper internal note' do
        post 'update_internal_note', {
          id: housing_request_true.id.to_s,
          internal_note: "test modified fuck!"
        }
        resp = JSON.parse(response.body)
        expect(resp['message']['value']).to eq("test modified ***")
      end
    end
    
    describe "POST /housing_requests/:id/update_entity_note" do
      let!(:user_collab) {create(:user, auth_roles: ['collaborateur offre logement'])}
      let!(:entity1) {create(:entity, siret: '00000000000001')}
      let!(:user2) {create(:user)}
      let!(:applicant42) {
        create(
          :applicant,
          user: user_collab,
          monthly_charges: {},
          nationality: "3",
          main_employer_siret: entity1.siret
        )
      }
      
      let!(:applicant2) {
        create(
          :applicant,
          user: user2,
          monthly_charges: {},
          nationality: "2"
        )
      }
      
      let!(:housing_request) {
        create(
          :housing_request,
          :with_dependents,
          is_active: true,
          main_applicant: applicant42,
          applicant_ids: [applicant2.id.to_s],
          entity_note: "test note"
        )
      }
      
      before do
        authenticated_header(request, user_collab)
        
        post 'update_entity_note', {
          id: housing_request.id.to_s,
          entity_note: "test modified"
        }
      end
      
      it 'returns the proper internal note' do
        resp = JSON.parse(response.body)
        expect(resp['message']['value']).to eq("test modified")
      end
    end
    
    describe 'POST /housing_requests/:id/update_from_sne' do
      let!(:user) {create(:user)}

      let!(:applicant) do
        create(
          :applicant, 
          user: user, 
          nationality: "1", 
          monthly_charges: {}, 
          first_name: "Martine", 
          last_name: "CHOPRON", 
          birth_date: Date.parse('1991-06-07'), 
          auto_set_user: false
        )
      end

      let(:dependents) do
        [
          {"last_name"=>"MOUZOPOULOS", "first_name"=>"Marie", "birth_date"=>"2011-03-07", "rel_type"=>"E", "sex"=>"F"},
          {"last_name"=>"BOUVET", "first_name"=>"Marie", "birth_date"=>"2019-03-20", "rel_type"=>"E", "sex"=>"F"}
        ]
      end

      let!(:housing_request) do
        create(:housing_request, user: user, is_active: true, main_applicant: applicant, dependents: dependents, seeked_localisation_list: [], motivations: ["accueillant", "super"])
      end

      let!(:nud) do
        create(
          :nud, 
          housing_request: housing_request, 
          department: nil, 
          creation_date: '2020-02-02T23:00:00', 
          unique_number: "1110121021492G1104"
        )
      end

      let!(:hri) do 
        create(
          :housing_request_initializer,
          done: false, 
          unique_number: nud.unique_number,
        )

      end 

      before do
        housing_request.reload
        authenticated_header(request, user)
      end

      it 'applicant_verified' do
        post 'update_from_sne', id: housing_request.id.to_s
        old_seeked_district_insee_codes   = housing_request.seeked_district_insee_codes
        old_monthly_revenues              = housing_request.monthly_revenues
        old_seeked_localisation_list      = housing_request.seeked_localisation_list
        housing_request.reload
        new_seeked_district_insee_codes   = housing_request.seeked_district_insee_codes
        new_monthly_revenues              = housing_request.monthly_revenues
        new_seeked_localisation_list      = housing_request.seeked_localisation_list
        
        expect(response.status).to eq(200)
        expect(old_seeked_district_insee_codes).not_to  eq(new_seeked_district_insee_codes)
        expect(old_monthly_revenues).not_to             eq(new_monthly_revenues)
        expect(old_seeked_localisation_list).not_to     eq(new_seeked_localisation_list)
      end

      context 'when not applicant_verified' do 
        before do
          housing_request.reload.update(applicant_verified: false)
          hri.update(hashed_pin_code: hri.hash_string("test"), pin_code_created_at: DateTime.now)
          authenticated_header(request, user)
        end

        it 'error when missing hri' do 
          hri.destroy
          post 'update_from_sne', id: housing_request.id.to_s
          expect(response.status).to eq(404)
        end 

        it 'error when no code' do 
          post 'update_from_sne', id: housing_request.id.to_s
          expect(response.status).to eq(422)
        end 

        it 'error when wrong pin_code' do 
          post 'update_from_sne', id: housing_request.id.to_s, pin_code: 'lol'
          expect(response.status).to eq(422)
        end 

        it 'works' do 
          post 'update_from_sne', id: housing_request.id.to_s, pin_code: 'test'
          expect(response.status).to eq(200)
          expect(housing_request.reload.applicant_verified).to eq(true)
        end 
      end 

      it 'when not active request' do 
        housing_request.set(is_active: false, updated_at: DateTime.now)
        post 'update_from_sne', id: housing_request.id.to_s
        expect(response.status).to eq(422)
        expect(json['errors'][0]['title']).to include('trying_to_update_an_inactive_request')
      end 

      # TODO cassette
      # before do
      #   authenticated_header(request, user)
      #   require 'vcr_config'
      #   VCR.use_cassette("update_from_sne", :match_requests_on => [:method, :uri, :path]) do
      #     housing_request.get_xml_from_sne(nil)
      #   end
      # end

      # it 'should update housing request from sne' do
      #   old_seeked_district_insee_codes   = housing_request.seeked_district_insee_codes
      #   old_monthly_revenues              = housing_request.monthly_revenues
      #   old_seeked_localisation_list      = housing_request.seeked_localisation_list

      #   post 'update_from_sne', id: housing_request.id.to_s

      #   expect(response.status).to eq(200)
      #   housing_request.reload

      #   new_seeked_district_insee_codes   = housing_request.seeked_district_insee_codes
      #   new_monthly_revenues              = housing_request.monthly_revenues
      #   new_seeked_localisation_list      = housing_request.seeked_localisation_list

      #   expect(old_seeked_district_insee_codes).not_to  eq(new_seeked_district_insee_codes)
      #   expect(old_monthly_revenues).not_to             eq(new_monthly_revenues)
      #   expect(old_seeked_localisation_list).not_to     eq(new_seeked_localisation_list)

      #   expect(housing_request.motivation1).to          eq('09')
      #   expect(housing_request.motivation2).to          be_nil
      #   expect(housing_request.motivation3).to          be_nil
      # end
    end

    describe "POST /housing_requests/:id/force_update_renewal_values" do
      let!(:created_at) { Date.today}
      let!(:applicant) { create(:applicant) }
      let!(:hr) {
        create(
          :housing_request,
          user: user,
          is_active: true,
          created_at: created_at,
          main_applicant: applicant,
          is_house_owner: true,
        )
      }
      let!(:nud) { create(:nud, housing_request: hr, unique_number: '1110121021492G1104')}

      before do
        authenticated_header(request, user)
      end

      it 'works' do
        hr.reload
        expect(hr.next_sne_renewal_date).to eq(created_at + 1.year)
        expect(hr.need_to_be_renewed).to eq(false)
        post 'force_update_renewal_values', id: hr.id
        hr.reload
        xml_sne_created_at = Date.new(2020, 01, 01)
        expect(hr.sne_created_at.to_date).to eq(xml_sne_created_at)
        expect(hr.next_sne_renewal_date).to eq(xml_sne_created_at + 1.year)
        expect(hr.need_to_be_renewed).to eq(true)
        expect(hr.is_house_owner).to eq(true)
        expect(response.status).to eq(200)
        expect(json['data']['attributes']['next_sne_renewal_date']).to eq((xml_sne_created_at + 1.year).to_s)
      end
    end

    describe 'POST /housing_requests/send_message' do
      let!(:collaborateur) {create :user, auth_roles: ['collaborateur offre logement']}
      subject {post 'send_message', to: 'test@test.com', data: {attributes: {title: 'title', body: 'body'}}}
      before do
        authenticated_header(request, collaborateur)
      end
      
      it 'creates a new Doc' do
        expect {subject}.to change {Doc.count}.by 1
        
        expect(response.status).to eq(201)
        expect(json['data']['attributes']['title']).to eq('title')
        expect(json['data']['attributes']['body']).to eq('body')
      end
      
      it 'sends an email' do
        expect(Agent).to receive_message_chain(:regular_mail, :deliver_later)
        subject
      end
    end
  end
  
  describe "GET /housing_requests/:id/xml_sopra_errors" do
    let!(:a) {create(:applicant)}
    let!(:hr) {create(:housing_request, main_applicant: a)}
    let!(:hr2) {create(:housing_request, main_applicant: a)}
    let!(:error1) {create(:xml_sopra_error, item_id: hr.id.to_s, item_class: hr.class, error_content: 'test1')}
    let!(:error2) {create(:xml_sopra_error, item_id: hr.id.to_s, item_class: hr.class, error_content: 'test2')}
    let!(:error3) {create(:xml_sopra_error, item_id: hr.id.to_s, item_class: hr.class, error_content: 'test3')}
    let!(:error4) {create(:xml_sopra_error, item_id: hr2.id.to_s, item_class: hr.class, error_content: 'test4')}
    
    before do
      authenticated_header(request, responsible)
      get 'xml_sopra_errors', {id: hr.id.to_s}
    end
    
    it "as valid data" do
      datas = json['data']
      expect(response.status).to eq(200)
      expect(datas.length).to eq(3)
    end
  
  end

  describe "GET /housing_requests/clp_details" do 
    let!(:applicant) {create(:applicant, :with_eligible_offers, offers_count: 7)}
    let!(:rh) {create(:reservation_holder, name: 'test', siren: ['1233456789'])}

    it 'works' do 
      authenticated_header(request, responsible)
      # update status offer
      Hermes::HousingOffer.each_with_index do |o, i|
        status = 'LOGEMENT PUBLIE'
        status = 'COMMERCIALISE SANS PUBLICATION' if i == 2
        status = 'EN COURS DE DESIGNATION' if i == 3 || i == 4
        status = 'ANNULEE' if i == 5
        o.update(offer_status: status)
      end 

      # get eligible_offers_nb_detailed
      get 'clp_details', { als_ids: [applicant.first_active_request.als_id]}
      expect(json['data'].first['attributes']['eligible_offers_count']).to eq(3)

      # check cache
      Hermes::HousingOffer.find_by(offer_status: 'LOGEMENT PUBLIE').set(offer_status: 'ANNULEE', published: false)
      get 'clp_details', { als_ids: [applicant.first_active_request.als_id]}
      expect(json['data'].first['attributes']['eligible_offers_count']).to eq(2)

      # check clear cache after 10sec
      Timecop.freeze(DateTime.current + 11.seconds) 
      get 'clp_details', { als_ids: [applicant.first_active_request.als_id]}
      expect(json['data'].first['attributes']['eligible_offers_count']).to eq(2)
      Timecop.return

      # check if the detailed_eligibility clear the chache
      @controller.instance_variable_set(:@options, {}) # clear the option from controller bc only one controller instance used in spec
      get 'show', { id: applicant.first_active_request.id}
      expect(json['data']['attributes']['eligible_offers_count'].present?).to eq(false)
    end 

    it 'with reserving_siren' do 
      authenticated_header(request, responsible)

      Hermes::HousingOffer.each_with_index do |o, i|
        o.set(reserving_siren: rh.siren.first, updated_at: DateTime.current)if i < 3
      end 

      get 'clp_details', { als_ids: [applicant.first_active_request.als_id], reserving_siren: rh.siren.first}
      expect(json['data'].first['attributes']['eligible_offers_count']).to eq(3)
    end 

    it 'with request missing eligibility' do 
      authenticated_header(request, responsible)
      applicant.first_active_request.set(eligibility_query_base: nil)
      get 'clp_details', { als_ids: [applicant.first_active_request.als_id]}
      expect(json['data'].first['attributes']['eligible_offers_count']).to eq(0)
    end 
  end

  describe "GET /housinng_requests/config_change_minus_two" do 
    before do
      authenticated_header(request, responsible)
    end 

    it 'without config' do 
      get 'config_change_minus_two'
      expect(json['change_minus_two_date']).to eq(nil)
      expect(json['change_minus_two_year']).to eq(nil)
    end 

    it 'without config' do 
      date = Date.new(2021, 4, 1,)
      HERMES_CONFIG['change_minus_two_date'] = date
      HERMES_CONFIG['change_minus_two_year'] = date.year - 1
      get 'config_change_minus_two'
      expect(json['change_minus_two_date']).to eq("2021-04-01")
      expect(json['change_minus_two_year']).to eq(2020)
      HERMES_CONFIG['change_minus_two_date'] = nil
      HERMES_CONFIG['change_minus_two_year'] = nil

    end
  end 

  describe "PUT /housing_requests/:id/update_next_renewal_date_set_by_sne" do 
    let!(:hr) {create(:active_request)}

    it 'unauthrorized for non superadmin' do 
      authenticated_header(request, user)
      put 'update_next_renewal_date_set_by_sne', {id: hr.id.to_s}
      expect(response).to have_http_status(401)
    end 

    it 'when missing hr' do 
      authenticated_header(request, superadministrateur)
      put 'update_next_renewal_date_set_by_sne', {id: "not_found"}
      expect(response).to have_http_status(404)
    end 

    it 'when missing params' do 
      authenticated_header(request, superadministrateur)
      put 'update_next_renewal_date_set_by_sne', {id: hr.id.to_s}
      expect(response).to have_http_status(422)
    end 

    it 'udpate' do 
      authenticated_header(request, superadministrateur)
      put 'update_next_renewal_date_set_by_sne', {id: hr.id.to_s, next_renewal_date_set_by_sne: '2022-01-01'}
      expect(response).to have_http_status(200)
      expect(json["data"]["attributes"]["next_renewal_date_set_by_sne"]).to eq('2022-01-01')
      expect(json["data"]["attributes"]["next_sne_renewal_date"]).to eq('2022-01-01')
    end 

  end 

  describe "PUT /housing_requests/:id/entity_validation" do 
    let!(:hr) {create(:complete_request)}
    let!(:entity) { create(:entity)}
    let!(:administrator) { create(:user)}
    let!(:user) { create(:user)}
    let!(:reader) { create(:user)}
    let!(:other_user) { create(:user)}
    let!(:ur_administrator) {create(:user_right, user: administrator, entity: entity, type: 'entity', right: 'administrator')}
    let!(:ur_user) {create(:user_right, user: user, entity: entity, type: 'entity', right: 'user')}
    let!(:ur_reader) {create(:user_right, user: reader, entity: entity, type: 'entity', right: 'reader')}
    let!(:params) { {id: hr.id, data: { attributes: { priority_points_indexes: [100], entity_validated: true, entity_refused: false }}}}

    before do 
      hr.set(entity_id: entity.id)
      authenticated_header(request, superadministrateur)
    end 

    it 'not found' do 
      put 'entity_validation', {id: 'lol'}
      expect(response).to have_http_status(404)
      expect(json['errors'].first['code']).to include('instance_not_found')
    end 

    it 'not_entity_found' do 
      hr.set(entity_id: nil)
      put 'entity_validation', params
      expect(response).to have_http_status(422)
      expect(json['errors'].first['code']).to include('not_entity_found')
    end 

    [
      :priority_points_indexes,
      :entity_validated,
      :entity_refused
    ].each do |field|
      it "missing params #{field}" do 
        attributes = params[:data][:attributes]
        put 'entity_validation', {id: hr.id, data: {attributes: attributes.except(field)}}
        expect(response).to have_http_status(422)
        expect(json['errors'].first['code']).to include("#{field}_param_missing")
      end
    end

    it 'not enought points' do 
      attributes = params[:data][:attributes]
      put 'entity_validation', {id: hr.id, data: {attributes: attributes.merge(priority_points_indexes: [1000000])}}
      expect(response).to have_http_status(422)
      expect(json['errors'].first['detail']).to include('No available points on the entity')
    end 

    [
      "administrator",
      "user",
      "superadministrateur"
    ].each do |uname|
      it "works for #{uname}" do 
        authenticated_header(request, eval(uname))
        put 'entity_validation', params
        expect(response).to have_http_status(200)
        expect(json['data']['attributes']['priority_points_indexes']).to eq([100])
      end 
    end

    it "not_authorized for only reader" do 
      authenticated_header(request, reader)
      put 'entity_validation', params
      expect(response).to have_http_status(401)
      expect(json['errors'].first['code']).to include('not_authorized')
    end 
  end
  
end

require_relative '../../swagger_helper'

describe Hermes::Api::Dmo::HousingRequestsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
  tags = 'API_DMO_HOUSING_REQUESTS'
  include JwtAuthHelper
  extend HermesSwaggerJsonParams
  include SwaggerJsonAttributes
  let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}
  let!(:user) {create(:user, first_name: 'Jackie', last_name: 'Mike', email: 'jackie.mike@gmail.com')}
  let!(:applicant) {create(:applicant, user_id: responsible.id.to_s, nationality: "1")}
  let!(:housing_channel) {create :housing_channel, key: 'dalo'}
  let!(:dc_channel) {create(:housing_channel, key: 'dc')}
  let!(:superadmin) {create(:user, auth_roles: ['superadministrateur'])}
  let!(:admin) {create(:administrator)}
  
  let!(:main_applicant) {
    create(
      :applicant,
      user: superadmin,
      monthly_charges: {},
      nationality: "3"
    )
  }

  let!(:user_applicant) {create(:user, auth_roles: ['collaborateur'])}
  let!(:entity) {create(:entity, siret: '00000000000001')}
  let!(:main_applicant) {create(:applicant, user: user_applicant, main_employer_siret: entity.siret)}
  let!(:housing_request) {create(:housing_request, user: user_applicant, main_applicant_id: main_applicant.id, is_active: true)}
  
  path '/api/dmo/housing_requests/admin_search' do
    get 'admin_search ' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      
      security [Bearer: {}]
      parameter name: :params, in: :body, schema: {
        type: :object,
        properties: {
          entity_params: {
            type:  :object,
            properties: {
              id: {type: :string},
              id_cristal: {type: :string},
              siren: {type: :string},
              siret: {type: :string},
              title: {type: :string},
            }
          },
          applicant_params: {
            type:  :object,
            properties: {
              main_applicant_id: { type: :string},
              requester_email: { type: :string},
              requester_firstname: { type: :string},
              requester_lastname: { type: :string},
            }
          },
          unique_numbers: {type: :string},
        },
      }
      let(:params) {{}}
      
      response '401', 'unauthorized' do
        let(:Authorization) {"Bearer"}

        run_test!
      end
      
      response '200', 'housing_request found' do
        let(:Authorization) {"Bearer #{token_for(superadmin)}"}
        let(:params) do
          {
            id: housing_request.id.to_s
          }
        end
        
        include_context 'with integration test'
        include_context 'not containings errors'
        
        after do |_example|
        end
      end
    end
  end

  ['index', 'search'].each do |type|
    path "/api/dmo/housing_requests#{type == 'search' ? '/search' : ''}" do
      method = type == 'search' ? :post : :get
      send(method, 'Retrieves all housing_requests') do
        tags tags
        produces 'application/json'
        Hermes::Api::Dmo::HousingRequestsController.acceptable_params.sort.each do |params_name, options|
          parameter name: params_name.to_s, in: options[:in] || :query, type: options[:type], required: options[:required] || false, description: options[:description] || ''
        end
        parameter name: 'postal_codes', in: :query, type: :array, required: false

        security [Bearer: {}]
        paginator

        response '401', 'unauthorized' do
          let(:Authorization) { "Bearer" }

          run_test!
        end 

        response '200', 'housing_offers found' do
          let(:Authorization) {"Bearer #{token_for(admin)}"}
          let!(:housing_requests) {create_list(:housing_request, 5, main_applicant_id: main_applicant.id, is_active: true)}

          include_context 'with integration test'
          include_context 'valid jsonapi collection' do
            let(:expected_collection) {housing_requests + [housing_request]}
          end
        end
      end
    end
  end 
    
  path '/api/dmo/housing_requests/{id}' do
    get 'Retrieves a single housing_request by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      let(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
      let(:id) {housing_request.id}

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }
        run_test!
      end 

      response '200', 'housing_request found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        
        include_context 'with integration test'
        include_context 'valid jsonapi ressource' do
          let(:expected_ressource) {housing_request}
        end
      end
    end
  end

  path '/api/dmo/housing_requests' do
    post 'Create housing_request' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::HousingRequestsController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter name: :params, in: :body, schema: jsonify_params(json_params)
      
      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )

      let(:params) do
        jsonapi_attributes(
          {main_applicant_id: main_applicant.id.to_s}
        )
      end

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '201', 'housing_request created' do
        let(:Authorization) {"Bearer #{token_for(responsible)}"}

        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'

        after do |_example|
          json = JSON.parse(response.body)

          expect(json['data']).to be_a Hash
          expect(json['data']['id']).to_not be nil
          expect(json['data']['type']).to eq 'housing_requests'
          expect(json['data']['attributes']['main_applicant_id']).to eq(main_applicant.id.to_s)
        end
      end
    end
  end

  path '/api/dmo/housing_requests/{id}' do
    put 'edit housing_request' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::HousingRequestsController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter name: :params, in: :body, schema: jsonify_params(json_params)
      let(:id) {housing_request.id}

      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )

      let(:params) do
        jsonapi_attributes(
          {main_applicant_id: main_applicant.id.to_s}
        )
      end
      
      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }
        run_test!
      end 

      response '200', 'housing_request updated' do
        let(:Authorization) {"Bearer #{token_for(responsible)}"}

        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'
        after do |_example|
          json = JSON.parse(response.body)

          expect(json['data']).to be_a Hash
          expect(json['data']['id']).to_not be nil
          expect(json['data']['type']).to eq 'housing_requests'
        end
      end
    end
  end

  path '/api/dmo/housing_requests/{id}' do
    delete 'Delete a single housing_request by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      let(:id) {create(:housing_request, user_id: responsible.id.to_s, main_applicant_id: applicant.id.to_s).id}

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }
        run_test!
      end 

      response '204', 'housing_request deleted' do
        let(:Authorization) {"Bearer #{token_for(responsible)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/clp_details' do
    get 'Retrieves housing requests with recount of eligible_offers_count' do
      tags tags
      produces 'application/json'
      parameter name: 'als_ids[]', in: :query, type: :array, items: {type: :string}, required: true, collectionFormat: :multi
      
      security [Bearer: {}]
      paginator
      let(:housing_requests) {create_list(:housing_request, 5, main_applicant_id: applicant.id.to_s)}
      let!(:'als_ids[]') { ['1', '2', '3', '4', '5'] }

      before do 
        housing_requests.each_with_index do |housing_request, i|
          housing_request.set(als_id: (i+1).to_s)
        end 
      end 

      response '200', 'housing_request found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        
        include_context 'with integration test'
        include_context 'valid jsonapi collection' do
          let(:expected_collection) {housing_requests}
        end
      end
    end
  end
  
  path '/api/dmo/housing_requests/config_change_minus_two' do
    get 'Get the configuration of year for the reference tax revenue ' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      
      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Configuration found' do
        let(:Authorization) {"Bearer #{token_for(user_applicant)}"}
        run_test!
      end
    end
  end
      
  path '/api/dmo/housing_requests/send_message' do
    post 'Send a message' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      
      parameter(
        name: :params,
        in: :body,
        schema: {
          name: {type: :string, required: false},
          title: {type: :string, required: false},
          to: {type: :string, required: false},
          cc: {type: :string, required: false},
          bcc: {type: :string, required: false},
          data: {
            type: :object,
            required: true,
            properties: {
              attributes: {
                type: :object,
                required: true,
                properties: {
                  title: {type: :string, required: false},
                  body: {type: :string, required: false},
                }
              }
            }
          }
        }
      )

      let(:params) do 
        {
          to: 'test@test.com', 
          data: {attributes: {title: 'title', body: 'body'}}
        }
      end

      let!(:collaborateur) { create(:user, auth_roles: ['collaborateur offre logement'])}

      security [Bearer: {}]
      json_params = {}
      
      response '201', 'message sent' do
        let(:Authorization) {"Bearer #{token_for(collaborateur)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/send_flux_thirty' do
    post 'Send flux 30' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      
      parameter name: 'als_ids[]', in: :query, type: :array, items: {type: :string}, required: true, collectionFormat: :multi
      let!(:'als_ids[]') { ['1'] }

      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Flux 30 sent' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/entity_note' do
    get 'Get entity_note' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      let(:id) {housing_request.id}

      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Entity note found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/internal_note' do
    get 'Get internal_note' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      let(:id) {housing_request.id}

      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Internal note found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/update_entity_note' do
    put 'Update entity_note' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      let(:id) {housing_request.id}
      parameter name: :entity_note, in: :query, type: :string
      let(:entity_note) {'test'}

      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Entity note updated' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/update_internal_note' do
    put 'Update internal_note' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      let(:id) {housing_request.id}
      parameter name: :internal_note, in: :query, type: :string
      let(:internal_note) {'test'}

      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Internal note updated' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/housing_request_folder' do
    get 'Get Folder' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      let(:id) {housing_request.id}

      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Folder found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/entity_requests' do
    get 'Get entity requests' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      let(:id) {housing_request.id}

      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Entity Request found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end

  ['waiting_points', 'spent_points'].each do |type|
    path "/api/dmo/housing_requests/{id}/#{type}" do
      get 'Get waiting_points' do
        tags tags
        produces 'application/json'
        consumes 'application/json'
        parameter name: :id, in: :path, type: :string
        let(:id) {housing_request.id}

        security [Bearer: {}]
        json_params = {}
        
        response '200', "#{type} found" do
          let(:Authorization) {"Bearer #{token_for(admin)}"}
          run_test!
        end
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/add_points_to_housing_request' do
    post 'Add points to housing_request' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string
      let(:id) {housing_request.id}

      parameter name: :params, in: :body, schema: {
        type: :object,
        properties: {
          entity_id: {type: :string, required: true},
          points: {type: :integer, required: true}
        }
      }

      let(:params) {{entity_id: entity.id, points: 10}}

      security [Bearer: {}]
      json_params = {}
      
      response '200', 'Point added' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end
  
  path '/api/dmo/housing_requests/{id}/add_housing_channel' do
    post 'Add a new key to the housing request housing_channel_pending_keys array' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string, required: true
      parameter name: :params, in: :body, schema: {
        type: :object,
        properties: {
          keys: { type: :array, items: { type: :string }, collectionFormat: :multi },
          required: [:keys]
        }
      }
      security [Bearer: {}]
      response '200', 'housing channel key added' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
        let(:id) {housing_request.id}
        let(:params) do
          {keys: ['dalo']}
        end
        
        include_context 'with integration test'
        include_context 'not containings errors'
        after do |_example|
          JSON.parse(response.body)
        end
      end
    end
  end
  
  path '/api/dmo/housing_requests/{id}/validate_housing_channel' do
    post 'Validate a pending key to the housing request housing_channel_keys array' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string, required: true
      parameter name: :params, in: :body, schema: {
        type: :object,
        properties: {
          keys: { type: :array, items: { type: :string }, collectionFormat: :multi },
          required: [:keys]
        }
      }
      security [Bearer: {}]
      response '200', 'housing channel key validated' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s, housing_channel_pending_keys: ['dalo'])}
        let(:id) {housing_request.id}
        let(:params) do
          {keys: ['dalo']}
        end
        
        include_context 'with integration test'
        include_context 'not containings errors'
        after do |_example|
          JSON.parse(response.body)
        end
      end
    end
  end
  
  path '/api/dmo/housing_requests/{id}/remove_housing_channel' do
    post 'Remove an housing channel' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      parameter name: :key, in: :query, type: :string, required: true
      security [Bearer: {}]
      let(:key) {housing_channel.key}

      paginator
      let(:id) {housing_request.id}

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '200', 'removed housing channel' do
        let(:'Content-Type') {'multipart/form-data'}
        let(:Authorization) {"Bearer #{token_for(responsible)}"}

        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/update_from_sne' do
    post 'update the info with ressources on SNE' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string, required: true
      security [Bearer: {}]
      let(:id) {housing_request.id}
      let(:Authorization) {"Bearer #{token_for(admin)}"}
      let!(:nud) do
        create(
          :nud, 
          housing_request: housing_request, 
          department: nil, 
          creation_date: '2020-02-02T23:00:00', 
          unique_number: "1110121021492G1104"
        )
      end

      response '200', 'info have been updated' do
        
        include_context 'with integration test'
        include_context 'not containings errors'
        after do |_example|
          
          json = JSON.parse(response.body)
        end
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/validate_code' do
    post 'Validate an entity_code for the main applicant' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string, required: true
      parameter name: :params, in: :body, schema: {
        type: :object,
        properties: {
          code: {type: :string},
          required: [:code]
        }
      }
      security [Bearer: {}]
      response '200', 'entity code validated for the main applicant' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:entity) {create :entity}
        let(:entity_code) {create :entity_code, entity_id: entity.id, applicant_id: applicant.id}
        let(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}
        let(:id) {housing_request.id}
        let(:params) do
          {code: entity_code.code}
        end
        
        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'
        after do |_example|
          JSON.parse(response.body)
        end
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/xml_sopra_errors' do
    get 'Get xml sopra errros' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      let(:id) {housing_request.id}

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }
        run_test!
      end 

      response '200', 'xml sopra error found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/force_update_renewal_values' do
    post 'force update of renewal_values' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string, required: true
      security [Bearer: {}]
      let(:id) {housing_request.id}
      let(:Authorization) {"Bearer #{token_for(admin)}"}
      let!(:nud) { create(:nud, housing_request: housing_request, unique_number: '1110121021492G1104')}


      response '200', 'renewal_values updated' do
        
        include_context 'with integration test'
        include_context 'not containings errors'
        after do |_example|
          
          json = JSON.parse(response.body)
        end
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/update_next_renewal_date_set_by_sne' do
    put 'update next_renewal_date_set_by_sne' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      parameter name: :id, in: :path, type: :string, required: true
      security [Bearer: {}]
      let(:id) {housing_request.id}
      let(:Authorization) {"Bearer #{token_for(admin)}"}
      let!(:nud) { create(:nud, housing_request: housing_request, unique_number: '1110121021492G1104')}
      parameter name: :next_renewal_date_set_by_sne, in: :query, type: :string, required: true

      let(:next_renewal_date_set_by_sne) { '2020-02-02T23:00:00'} 

      response '200', 'renewal_values updated' do
        
        include_context 'with integration test'
        include_context 'not containings errors'
        after do |_example|
          
          json = JSON.parse(response.body)
        end
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/eligible_offers' do
    get 'Retrieves eligible_offers relative to a given housing_request' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]
      Hermes::Api::Dmo::HousingOffersController.acceptable_params.each do |acceptable_param, options|
        parameter name: acceptable_param.to_s, in: :query, required: false, type: options[:type]
      end
      paginator
      let(:id) { housing_request.id }

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '200', 'eligible_offers found' do
        let(:Authorization) {"Bearer #{token_for(responsible)}"}

        include_context 'not containings errors'
        run_test!
      end
    end
  end

  path '/api/dmo/housing_requests/{id}/is_eligible_offer/{offer_id}' do
    get 'Check if a request is eligible to an request' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      parameter name: :offer_id, in: :path, type: :string

      security [Bearer: {}]

      paginator
      let(:id) {housing_request.id}
      let(:offer_id) { create(:housing_offer).id }

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        run_test!
      end 

      response '200', 'is eliglibe' do
        let(:'Content-Type') {'multipart/form-data'}
        let(:Authorization) {"Bearer #{token_for(responsible)}"}

        run_test!
      end
    end
  end
end
