require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

THRESHOLD_MAP = {
  'Paris et alentours' => {
    'thresholds_list' => [
      '12673',
      '14679'
    ],
    'another_person' => '6723'
  }
}.freeze

describe Hermes::Api::Dmo::ThresholdsController, :type => :api do
  routes {Hermes::Engine.routes}
  include Devise::Test::ControllerHelpers
  include JwtAuthHelper
  let!(:responsible) { create(:user, roles: ['Administrator']) }
  let!(:user) { create(:user) }

  #TESTS BASIQUES D'API

  describe 'GET /thresholds' do
    describe 'GET /thresholds' do
      let!(:threshold1) {create(:threshold, user_id: responsible.id.to_s, thresholds_map: THRESHOLD_MAP, type: 'ceils')}
      let!(:threshold3) {create(:threshold, user_id: responsible.id.to_s, thresholds_map: THRESHOLD_MAP, type: 'ceils')}
      let!(:threshold2) {create(:threshold, user_id: responsible.id.to_s, thresholds_map: THRESHOLD_MAP, type: 'ceils')}
      before :each do
        authenticated_header(request, responsible)
        get 'index'
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
        expect(json['data'].length).to eq(3)
        expect(json['meta']['pagination']).to be_present
        expect(json['meta']['pagination']['total_objects']).to eq(3)
      end
    end

    describe 'GET /thresholds with operators' do
      let!(:threshold1) {create(:threshold, user_id: responsible.id.to_s, thresholds_map: THRESHOLD_MAP, type: 'acceuillant')}
      let!(:threshold3) {create(:threshold, user_id: responsible.id.to_s, thresholds_map: THRESHOLD_MAP, type: 'ceils')}
      let!(:threshold2) {create(:threshold, user_id: responsible.id.to_s, thresholds_map: THRESHOLD_MAP, type: 'Seils')}

      describe 'with $in' do
        before :each do
          authenticated_header(request, responsible)
          get 'index', {type: {'$in': ['acceuillant', 'flamand']}}
        end
        it 'returns results' do
          expect(response.status).to eq(200)
          expect(json['meta']['pagination']['total_objects']).to eq(1)
        end
      end
      describe 'with $ne' do
        before :each do
          authenticated_header(request, responsible)
          get 'index', {type: {'$ne': 'acceuillant'}}
        end
        it 'returns sorted results' do
          expect(response.status).to eq(200)
          expect(json['meta']['pagination']['total_objects']).to eq(2)
        end
      end
      describe 'with $regex' do
        before :each do
          authenticated_header(request, responsible)
          get 'index', {type: {'$regex': /^S/}}
        end
        it 'returns sorted results' do
          expect(response.status).to eq(200)
          expect(json['meta']['pagination']['total_objects']).to eq(1)
        end
      end
    end
  end

  describe 'GET /thresholds/:id' do
    describe 'verif it work' do
      let!(:threshold1) {create(:threshold, user_id: responsible.id.to_s, thresholds_map: THRESHOLD_MAP, type: 'acceuillant')}
      before :each do
        authenticated_header(request, responsible)
        get 'show', {id: threshold1.id}
      end
      it 'returns status code 200' do
        expect(response).to have_http_status(200)
      end
      it 'returns a valid json' do
        expect(json).not_to be_empty
      end
    end
    describe 'verif it work not work with a non authorized user' do
      let!(:threshold1) {create(:threshold, user_id: responsible.id.to_s, thresholds_map: THRESHOLD_MAP, type: 'acceuillant')}

      before :each do
        authenticated_header(request, user)
        get 'show', {id: threshold1.id}
      end
      it 'returns status code 404' do
        expect(response).to have_http_status(404)
      end
    end
  end
end

require_relative '../../swagger_helper'

describe Hermes::Api::Dmo::ThresholdsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
  tags = 'API_DMO_THRESHOLDS'
  include JwtAuthHelper
  extend HermesSwaggerJsonParams
  include SwaggerJsonAttributes
  let!(:admin) { create(:user, roles: ['Administrator']) }

  path '/api/dmo/thresholds' do
    get 'Retrieves all thresholds' do
      tags tags
      produces 'application/json'

      Hermes::Api::Dmo::ThresholdsController.acceptable_params.sort.each do |params_name, options|
        parameter name: params_name.to_s, in: options[:in] || :query, type: options[:type], required: options[:required] || false, description: options[:description] || ''
      end

      security [Bearer: {}]
      paginator

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }

        run_test!
      end 

      response '200', 'thresholds found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let!(:thresholds) { create_list(:threshold, 5, type: 'test', thresholds_map: THRESHOLD_MAP) }

        include_context 'with integration test'
        include_context 'valid jsonapi collection' do
          let(:expected_collection) { thresholds }
        end
      end
    end
  end

  path '/api/dmo/thresholds/{id}' do
    get 'Retrieves a single threshold by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        let(:id) { 1 }

        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }

        run_test!
      end 
      
      response '200', 'threshold found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:threshold) {create(:threshold, type: 'test', thresholds_map: THRESHOLD_MAP)}
        let(:id) {threshold.id}

        include_context 'with integration test'
        include_context 'valid jsonapi ressource' do
          let(:expected_ressource) {threshold}
        end
      end
    end
  end

end
