require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

describe Hermes::Api::Dmo::UserFiltersController, :type => :api do
  routes {Hermes::Engine.routes}
  include Devise::Test::ControllerHelpers
  include JwtAuthHelper

  let!(:user1) { create(:user) }
  let!(:user2) { create(:user) }
  

  describe 'POST /user_filters' do
    let!(:params) { { data: { attributes: { user_id: user1.id.to_s, key: 'key1', value: {test: 'bob', a: []} } } } }
    subject { post 'create', params }

    before :each do 
      authenticated_header(request, user1)
    end 
    
    it 'missing params' do 
      params[:data][:attributes][:user_id] = nil
      subject
      expect(response).to have_http_status(422)
      expect(json['errors'].first['title']).to match(/missing_params/)
    end 

    it "error when trying to create for other user" do 
      params[:data][:attributes][:user_id] = user2.id
      subject
      expect(response).to have_http_status(401)
      expect(json['errors'].first['title']).to match(/cannot_create_filter_for_other_user/)
    end 
    
    it 'create basic one' do
      subject
      expect(response).to have_http_status(201)
      expect(json['data']['attributes']['key']).to eq('key1')
      expect(json['data']['attributes']['value']["test"]).to eq('bob')
      expect(json['data']['attributes']['value']["a"]).to eq([])
    end

    it 'update existing key' do 
      uf = Hermes::UserFilter.create(user: user1, key: 'key1', value: { test: 'john'})
      subject
      expect(response).to have_http_status(201)
      expect(json['data']['attributes']['key']).to eq('key1')
      expect(json['data']['attributes']['value']["test"]).to eq('bob')
      expect(json['data']['id']).to eq(uf.id.to_s)
    end 
  end

  describe 'GET /user_filters' do
    let!(:uf1) { create(:user_filter, user: user1, key: 'key1') }
    let!(:uf2) { create(:user_filter, user: user1, key: 'key2') }
    let!(:uf3) { create(:user_filter, user: user1, key: 'key3') }
    let!(:uf4) { create(:user_filter, user: user2, key: 'key1') }

    it 'only return user filters' do
      authenticated_header(request, user1)
      get 'index'
      expect(response).to have_http_status(200)
      expect(json['data'].length).to eq(3)
    end
  end 

  describe 'GET /user_filters/:id' do
    let!(:uf1) { create(:user_filter, user: user1, key: 'key1') }
    
    it 'error not found' do
      authenticated_header(request, user1)
      get 'show', {id: 'unknown'}
      expect(response).to have_http_status(404)
      expect(json['errors'].first['title']).to match(/instance_not_found/)
    end

    it 'error when not user' do
      authenticated_header(request, user2)
      get 'show', {id: uf1.id}
      expect(response).to have_http_status(401)
      expect(json['errors'].first['title']).to match(/cannot_see_this_user_filter/)
    end

    it 'works' do
      authenticated_header(request, user1)
      get 'show', {id: uf1.id}
      expect(response).to have_http_status(200)
    end

  end

  describe 'DELETE /user_filters/:id' do
    let!(:uf1) { create(:user_filter, user: user1, key: 'key1') }
    
    it 'error not found' do
      authenticated_header(request, user1)
      delete 'destroy', {id: 'unknown'}
      expect(response).to have_http_status(404)
      expect(json['errors'].first['title']).to match(/instance_not_found/)
    end

    it 'error when not user' do
      authenticated_header(request, user2)
      delete 'destroy', {id: uf1.id}
      expect(response).to have_http_status(401)
      expect(json['errors'].first['title']).to match(/cannot_delete_other_user_filter/)
    end

    it 'works' do
      authenticated_header(request, user1)
      delete 'destroy', {id: uf1.id}
      expect(response).to have_http_status(204)
    end
  end

end

require_relative '../../swagger_helper'

describe Hermes::Api::Dmo::UserFiltersController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
  tags = 'API_DMO_USER_FILTERS'
  include JwtAuthHelper
  extend HermesSwaggerJsonParams
  include SwaggerJsonAttributes
  let!(:admin) { create(:user, roles: ['Administrator']) }

  path '/api/dmo/user_filters' do
    get 'Retrieves all user_filters' do
      tags tags
      produces 'application/json'

      Hermes::Api::Dmo::UserFiltersController.acceptable_params.sort.each do |params_name, options|
        parameter name: params_name.to_s, in: options[:in] || :query, type: options[:type], required: options[:required] || false, description: options[:description] || ''
      end

      security [Bearer: {}]
      paginator

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }

        run_test!
      end 

      response '200', 'user_filters found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let!(:user_filters) { create_list(:user_filter, 1, key: 'test', user: admin) }

        include_context 'with integration test'
        include_context 'valid jsonapi collection' do
          let(:expected_collection) { user_filters }
        end
      end
    end
  end

  path '/api/dmo/user_filters' do
    post 'Create user_filter' do
      tags tags
      produces 'application/json'
      consumes 'application/json'
      security [Bearer: {}]
      json_params = {}
      Hermes::Api::Dmo::UserFiltersController.mutate_params.each do |key, options|
        json_params[key] = nice_type(options[:type])
      end
      parameter(
        name: :params,
        in: :body,
        schema: jsonify_params(json_params),
      )
      let(:params) do
        jsonapi_attributes(
          {  key: 'test', user_id: admin.id }
        )
      end

      response '201', 'user_filter created' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }

        include_context 'with integration test'
        include_context 'valid jsonapi'
        include_context 'not containings errors'

        after do |_example|
          json = JSON.parse(response.body)

          expect(json['data']).to be_a Hash
          expect(json['data']['id']).to_not be nil
          expect(json['data']['type']).to eq 'user_filters'
        end
      end
    end
  end

  path '/api/dmo/user_filters/{id}' do
    get 'Retrieves a single user_filter by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]

      response '401', 'unauthorized' do
        let(:Authorization) { "Bearer" }
        let(:id) { 1 }

        run_test!
      end 

      response '404', 'not found' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { 1 }

        run_test!
      end 
      
      response '200', 'user_filter found' do
        let(:Authorization) {"Bearer #{token_for(admin)}"}
        let(:user_filter) {create(:user_filter, key: 'test', user: admin)}
        let(:id) {user_filter.id}

        include_context 'with integration test'
        include_context 'valid jsonapi ressource' do
          let(:expected_ressource) {user_filter}
        end
      end
    end
  end

  path '/api/dmo/user_filters/{id}' do
    delete 'Delete a single user_filter by id' do
      tags tags
      produces 'application/json'
      parameter name: :id, in: :path, type: :string
      security [Bearer: {}]

      response '204', 'user_filter deleted' do
        let(:Authorization) { "Bearer #{token_for(admin)}" }
        let(:id) { create(:user_filter, key: 'test', user: admin).id }
        run_test!
      end
    end
  end

end
