# frozen_string_literal: true

require 'rails_helper'

RSpec.describe Hermes::Entity, type: :model do
  it 'has a valid factory' do
    entity = FactoryBot.build(:entity)
    expect(entity).to be_valid
    expect(entity.auto_validation).to be(false)
  end

  context 'with basic entity' do
    let!(:basic_entity) { create(:entity) }

    describe 'triggered side effects are valid' do
      it 'has a group with valid name' do
        expect(basic_entity.group.name).to eq "entity_#{basic_entity.id}"
        expect(basic_entity.readers).to include "g_#{basic_entity.group.id}"
      end
    end
  end

  context 'with essential items for priority points' do
    let!(:entity1) { create(:entity, siret: '12300000000000') }

    it 'has entity1 with annual stock 8500' do
      expect(entity1.annual_stock).to be 8500
    end

    describe 'entity1 priority points attribution' do
      single_request_attributed_points = [30, 40]
      request_nb = 2

      request_nb.times do |i|
        let!(:"applicant#{i}") { create(:applicant, main_employer_siret: entity1.siret,) }
        let!(:"request#{i}") do
          create(
            :housing_request,
            is_active: true,
            main_applicant: send("applicant#{i}"),
          )
        end
      end

      it 'updates entity1 points distribution' do
        [request0, request1].each do |hr|
          hr.update(
            entity_validated: true,
            priority_points_indexes: single_request_attributed_points,
          )
        end 
        spent_points = request_nb * single_request_attributed_points.sum
        r_entity1 = entity1.reload
        expect(r_entity1.total_spent).to be(spent_points)
        expect(r_entity1.spent_on_ongoing_requests).to be(spent_points)
        expect(r_entity1.available_points).to be(8500 - spent_points)
        expect(r_entity1.spent_on_completed_requests).to be(0).or(be nil)
        expect(r_entity1.spent_on_other).to be(0).or(be nil)
      end
  
      it 'update_points when annual stock changed' do
        expect(entity1.reload.annual_stock).to be 8500
  
        request0.update(
          entity_validated: true,
          priority_points_indexes: [1000],
        )
  
        expect(entity1.reload.annual_stock).to be 8500
        expect(entity1.reload.total_spent).to be 1000
        expect(entity1.reload.available_points).to be 7500
  
        entity1.update(
          annual_stock: 10000,
        )
        expect(entity1.reload.total_spent).to be 1000
        expect(entity1.reload.available_points).to be 9000
      end 
    end

  end

  describe 'clean_integer' do 
    let!(:entity) {create(:entity) }

    Hermes::Entity.not_null_integer_fields.each do |field|
      it "#{field} cannot be changed to nil" do 
        entity.update(field => nil)
        expect(entity.reload[field]).not_to eq(nil)
      end 

      it "#{field} be nil and update_points not throw error" do 
        entity.set(field => nil)
        entity.update_points
        expect(entity.reload[field]).to eq(nil)
      end 
    end 
  end 
end
